# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: hypernets/dispatchers/process/grpc/proto/proc.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='hypernets/dispatchers/process/grpc/proto/proc.proto',
  package='hypernets.dispatchers.process.grpc.proto',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n3hypernets/dispatchers/process/grpc/proto/proc.proto\x12(hypernets.dispatchers.process.grpc.proto\"c\n\x0eProcessRequest\x12\x0f\n\x07program\x18\x01 \x01(\t\x12\x0c\n\x04\x61rgs\x18\x02 \x03(\t\x12\x0b\n\x03\x63wd\x18\x03 \x01(\t\x12\x13\n\x0b\x62uffer_size\x18\x04 \x01(\x05\x12\x10\n\x08\x65ncoding\x18\x05 \x01(\t\"T\n\x0f\x44ownloadRequest\x12\x0c\n\x04peer\x18\x01 \x01(\t\x12\x0c\n\x04path\x18\x02 \x01(\t\x12\x13\n\x0b\x62uffer_size\x18\x03 \x01(\x05\x12\x10\n\x08\x65ncoding\x18\x04 \x01(\t\"\xae\x01\n\tDataChunk\x12J\n\x04kind\x18\x01 \x01(\x0e\x32<.hypernets.dispatchers.process.grpc.proto.DataChunk.DataKind\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\"G\n\x08\x44\x61taKind\x12\x06\n\x02IN\x10\x00\x12\x07\n\x03OUT\x10\x01\x12\x07\n\x03\x45RR\x10\x02\x12\x08\n\x04\x44\x41TA\x10\n\x12\x07\n\x03\x45ND\x10\x63\x12\x0e\n\tEXCEPTION\x10\x90\x03\x32\x8b\x02\n\rProcessBroker\x12z\n\x03run\x12\x38.hypernets.dispatchers.process.grpc.proto.ProcessRequest\x1a\x33.hypernets.dispatchers.process.grpc.proto.DataChunk\"\x00(\x01\x30\x01\x12~\n\x08\x64ownload\x12\x39.hypernets.dispatchers.process.grpc.proto.DownloadRequest\x1a\x33.hypernets.dispatchers.process.grpc.proto.DataChunk\"\x00\x30\x01\x62\x06proto3')
)



_DATACHUNK_DATAKIND = _descriptor.EnumDescriptor(
  name='DataKind',
  full_name='hypernets.dispatchers.process.grpc.proto.DataChunk.DataKind',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='IN', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='OUT', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERR', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DATA', index=3, number=10,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='END', index=4, number=99,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXCEPTION', index=5, number=400,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=388,
  serialized_end=459,
)
_sym_db.RegisterEnumDescriptor(_DATACHUNK_DATAKIND)


_PROCESSREQUEST = _descriptor.Descriptor(
  name='ProcessRequest',
  full_name='hypernets.dispatchers.process.grpc.proto.ProcessRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='program', full_name='hypernets.dispatchers.process.grpc.proto.ProcessRequest.program', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='args', full_name='hypernets.dispatchers.process.grpc.proto.ProcessRequest.args', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='cwd', full_name='hypernets.dispatchers.process.grpc.proto.ProcessRequest.cwd', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='buffer_size', full_name='hypernets.dispatchers.process.grpc.proto.ProcessRequest.buffer_size', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='encoding', full_name='hypernets.dispatchers.process.grpc.proto.ProcessRequest.encoding', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=97,
  serialized_end=196,
)


_DOWNLOADREQUEST = _descriptor.Descriptor(
  name='DownloadRequest',
  full_name='hypernets.dispatchers.process.grpc.proto.DownloadRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='peer', full_name='hypernets.dispatchers.process.grpc.proto.DownloadRequest.peer', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='path', full_name='hypernets.dispatchers.process.grpc.proto.DownloadRequest.path', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='buffer_size', full_name='hypernets.dispatchers.process.grpc.proto.DownloadRequest.buffer_size', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='encoding', full_name='hypernets.dispatchers.process.grpc.proto.DownloadRequest.encoding', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=198,
  serialized_end=282,
)


_DATACHUNK = _descriptor.Descriptor(
  name='DataChunk',
  full_name='hypernets.dispatchers.process.grpc.proto.DataChunk',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='kind', full_name='hypernets.dispatchers.process.grpc.proto.DataChunk.kind', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data', full_name='hypernets.dispatchers.process.grpc.proto.DataChunk.data', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _DATACHUNK_DATAKIND,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=285,
  serialized_end=459,
)

_DATACHUNK.fields_by_name['kind'].enum_type = _DATACHUNK_DATAKIND
_DATACHUNK_DATAKIND.containing_type = _DATACHUNK
DESCRIPTOR.message_types_by_name['ProcessRequest'] = _PROCESSREQUEST
DESCRIPTOR.message_types_by_name['DownloadRequest'] = _DOWNLOADREQUEST
DESCRIPTOR.message_types_by_name['DataChunk'] = _DATACHUNK
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ProcessRequest = _reflection.GeneratedProtocolMessageType('ProcessRequest', (_message.Message,), {
  'DESCRIPTOR' : _PROCESSREQUEST,
  '__module__' : 'hypernets.dispatchers.process.grpc.proto.proc_pb2'
  # @@protoc_insertion_point(class_scope:hypernets.dispatchers.process.grpc.proto.ProcessRequest)
  })
_sym_db.RegisterMessage(ProcessRequest)

DownloadRequest = _reflection.GeneratedProtocolMessageType('DownloadRequest', (_message.Message,), {
  'DESCRIPTOR' : _DOWNLOADREQUEST,
  '__module__' : 'hypernets.dispatchers.process.grpc.proto.proc_pb2'
  # @@protoc_insertion_point(class_scope:hypernets.dispatchers.process.grpc.proto.DownloadRequest)
  })
_sym_db.RegisterMessage(DownloadRequest)

DataChunk = _reflection.GeneratedProtocolMessageType('DataChunk', (_message.Message,), {
  'DESCRIPTOR' : _DATACHUNK,
  '__module__' : 'hypernets.dispatchers.process.grpc.proto.proc_pb2'
  # @@protoc_insertion_point(class_scope:hypernets.dispatchers.process.grpc.proto.DataChunk)
  })
_sym_db.RegisterMessage(DataChunk)



_PROCESSBROKER = _descriptor.ServiceDescriptor(
  name='ProcessBroker',
  full_name='hypernets.dispatchers.process.grpc.proto.ProcessBroker',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=462,
  serialized_end=729,
  methods=[
  _descriptor.MethodDescriptor(
    name='run',
    full_name='hypernets.dispatchers.process.grpc.proto.ProcessBroker.run',
    index=0,
    containing_service=None,
    input_type=_PROCESSREQUEST,
    output_type=_DATACHUNK,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='download',
    full_name='hypernets.dispatchers.process.grpc.proto.ProcessBroker.download',
    index=1,
    containing_service=None,
    input_type=_DOWNLOADREQUEST,
    output_type=_DATACHUNK,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_PROCESSBROKER)

DESCRIPTOR.services_by_name['ProcessBroker'] = _PROCESSBROKER

# @@protoc_insertion_point(module_scope)



# -*- coding:utf-8 -*-
"""

"""
import numpy as np

from hypernets.core.ops import Choice, Bool, Identity
from hypernets.core.search_space import HyperSpace, Int, Real
from hypernets.searchers.evolution_searcher import EvolutionSearcher
from hypernets.searchers.mcts_searcher import MCTSSearcher
from hypernets.searchers.random_searcher import RandomSearcher


def get_space():
    space = HyperSpace()
    with space.as_default():
        p1 = Int(1, 100)
        p2 = Choice(['a', 'b', 'c'])
        p3 = Bool()
        p4 = Real(0.0, 1.0)
        id1 = Identity(p1=p1)
        id2 = Identity(p2=p2)(id1)
        id3 = Identity(p3=p3)(id2)
        id4 = Identity(p4=p4)(id3)
    return space


def run_search():
    searchers = (
        RandomSearcher(get_space, space_sample_validation_fn=lambda s: True),
        MCTSSearcher(get_space, max_node_space=10),
        EvolutionSearcher(get_space, 5, 3, regularized=False)
    )

    for searcher in searchers:
        for i in range(100):
            space_sample = searcher.sample()
            assert space_sample.all_assigned == True
            print(searcher.__class__.__name__, i, space_sample.params_summary())
            searcher.update_result(space_sample, np.random.uniform(0.1, 0.9))


if __name__ == '__main__':
    run_search()



# -*- coding:utf-8 -*-
"""

"""
from collections import OrderedDict


class MutableScope:
    def __init__(self):
        self.reset()
        self.stack = []

    @property
    def current_path(self):
        return '.'.join(self.stack)

    def entry(self, name):
        self.stack.append(name)

    def exit(self):
        self.stack.pop()

    def reset(self):
        self.id_dict = OrderedDict()
        self.name_dict = OrderedDict()

    def register(self, mutable):
        assert isinstance(mutable, Mutable)

        if mutable.name is None:
            mutable.id = self.assign_id(mutable)
            mutable.name = mutable.id
        else:
            if self.name_dict.get(mutable.name) is not None:
                raise ValueError(f'name `{mutable.name}` is duplicate.')
            mutable.id = f'ID_{mutable.name}'

        self.name_dict[mutable.name] = mutable
        self.id_dict[mutable.id] = mutable

    def assign_id(self, mutable):
        prefix = mutable.__class__.__name__
        if mutable.type is not None:
            prefix = mutable.type + '_' + prefix
        i = 1
        while True:
            id = f'{prefix}_{i}'
            if id not in self.id_dict:
                break
            i += 1
        return id

    def get_mutable(self, id):
        return self.id_dict[id]

    def get_mutable_by_name(self, name):
        return self.name_dict[name]


class Mutable(object):
    def __init__(self, scope, name=None):
        self.attach_to_scope(scope, name)

    def attach_to_scope(self, scope, name=None):
        assert scope is not None, 'scope cannot be None'
        self.scope = scope
        self.name = name
        self.alias = None
        self.scope.register(self)
        self.path = scope.current_path

    def __repr__(self):
        # if self.alias is not None:
        #     return 'ALIAS:' + self.alias
        # else:
        #     return 'ID:' + self._id
        return self._id

    @property
    def type(self):
        return None

    @property
    def id(self):
        return self._id

    @id.setter
    def id(self, id):
        self._id = id

    def update(self):
        pass



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
from collections import defaultdict

import numpy as np
from scipy.cluster import hierarchy
from scipy.stats import spearmanr
from sklearn.metrics import get_scorer
from sklearn.model_selection import train_test_split

from hypernets.core import set_random_state, get_random_state
from hypernets.examples.plain_model import train
from hypernets.tabular.datasets import dsutils
from hypernets.tabular.feature_importance import permutation_importance_batch
from hypernets.tabular.sklearn_ex import MultiLabelEncoder


class Test_PermutationImportance():

    def test_collinear(self):
        df = dsutils.load_bank().head(10000)
        y = df.pop('y')
        df.drop(['id'], axis=1, inplace=True)
        corr = spearmanr(df).correlation
        corr_linkage = hierarchy.ward(corr)
        # fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 8))
        # dendro = hierarchy.dendrogram(
        #     corr_linkage, labels=df.columns.to_list(), ax=ax1, leaf_rotation=90
        # )
        # dendro_idx = np.arange(0, len(dendro['ivl']))
        # ax2.imshow(corr[dendro['leaves'], :][:, dendro['leaves']])
        # ax2.set_xticks(dendro_idx)
        # ax2.set_yticks(dendro_idx)
        # ax2.set_xticklabels(dendro['ivl'], rotation='vertical')
        # ax2.set_yticklabels(dendro['ivl'])
        # fig.tight_layout()
        # plt.show()

        cluster_ids = hierarchy.fcluster(corr_linkage, 1, criterion='distance')
        cluster_id_to_feature_ids = defaultdict(list)
        for idx, cluster_id in enumerate(cluster_ids):
            cluster_id_to_feature_ids[cluster_id].append(idx)
        selected_features = [df.columns[v[0]] for v in cluster_id_to_feature_ids.values()]
        assert selected_features == ['age', 'job', 'marital', 'education', 'default', 'balance', 'housing', 'loan',
                                     'contact', 'day', 'month', 'duration', 'campaign', 'pdays', 'poutcome']

    def test_basic(self):
        set_random_state(9527)
        df = dsutils.load_bank().head(3000)
        encoder = MultiLabelEncoder()
        df = encoder.fit_transform(df)
        y = df.pop('y')
        df.drop(['id'], axis=1, inplace=True)
        X_train, X_test, y_train, y_test = train_test_split(df, y, test_size=0.3, random_state=42)

        hm, _ = train(X_train, y_train, X_test, y_test, max_trials=5)

        best_trials = hm.get_top_trials(3)
        estimators = [hm.load_estimator(trial.model_file) for trial in best_trials]

        importances = permutation_importance_batch(estimators, X_test, y_test, get_scorer('roc_auc_ovr'), n_jobs=1,
                                                   n_repeats=5, random_state=get_random_state())

        feature_index = np.argwhere(importances.importances_mean < 1e-5)
        selected_features = [feat for i, feat in enumerate(X_train.columns.to_list()) if i not in feature_index]
        unselected_features = [c for c in X_train.columns.to_list() if c not in selected_features]
        set_random_state(None)

        print('selected:  ', selected_features)
        print('unselected:', unselected_features)
        # assert selected_features == ['job', 'marital', 'education', 'balance', 'housing', 'loan', 'contact', 'day',
        #                              'duration', 'campaign', 'pdays', 'previous', 'poutcome']
        # assert unselected_features == ['age', 'default', 'month']



import grpc

from hypernets.dispatchers.predict.grpc.proto import predict_pb2_grpc
from hypernets.dispatchers.predict.grpc.proto.predict_pb2 import PredictRequest
from hypernets.utils import logging

logger = logging.get_logger(__name__)


class PredictClient(object):

    def __init__(self, server):
        super(PredictClient, self).__init__()
        self.channel = grpc.insecure_channel(server)
        self.stub = predict_pb2_grpc.PredictServiceStub(self.channel)

        self.server = server
        self._closed = False

    def __del__(self):
        self.close()

    def close(self):
        if not self._closed:
            self.channel.close()

    def predict(self, data_file, result_file):
        try:
            request = PredictRequest(data_file=data_file, result_file=result_file)
            response = self.stub.predict(request)
            code = response.code
            return code
        except Exception as e:
            import traceback
            msg = f'[Predict {self.server}] {e.__class__.__name__}:\n'
            logger.error(msg + traceback.format_exc())

            return 98 if isinstance(e, grpc.RpcError) else 99



# -*- coding:utf-8 -*-
"""

"""
import numpy as np
import math
from collections import OrderedDict
from ..utils.common import generate_id
from ..core import get_random_state


class MCNode(object):
    def __init__(self, id, name, param_sample, parent=None, tree=None, is_terminal=False, random_state=None):
        self.visits = 0
        self.reward = 0.0
        self.rewards = []
        self.simulation_rewards = []

        assert name is not None
        if name != 'ROOT':
            assert param_sample.assigned
        self.name = name
        self.param_sample = param_sample
        self.parent = parent
        if parent is not None:
            assert parent.tree is not None
            self.tree = parent.tree
        else:
            assert tree is not None
            self.tree = tree
        self._is_terminal = is_terminal
        self._children = []
        self.id = id
        self.random_state = random_state if random_state is not None else get_random_state()

    @property
    def is_terminal(self):
        return self._is_terminal

    def set_terminal(self):
        self._is_terminal = True
        return self._is_terminal

    @property
    def expanded(self):
        return all([n.visits > 0 for n in self.children])

    @property
    def is_leaf(self):
        return len(self._children) <= 0

    @property
    def children(self):
        return self._children

    @property
    def depth(self):
        d = 0
        node = self.parent
        while node is not None:
            d += 1
            node = node.parent
        return d

    def add_child(self, param_sample):
        self.children.append(MCNode(generate_id(), param_sample.label, param_sample, parent=self))

    def set_parent(self, parent):
        assert self.parent is not None
        self.parent = parent

    def expansion(self, param_space, max_space):
        # print(f'Node expanssion: param_space:{param_space.label}')
        assert not param_space.assigned
        samples = param_space.expansion(sample_num=max_space)
        for param_sample in samples:
            self.add_child(param_sample)

    def random_sample(self):
        child = self.random_state.choice(self._children)
        return child

    def info(self):
        return f'Reward:{self.reward}, Visits:{self.visits}, Name:{self.name}, Alias:{self.param_sample.alias if self.param_sample is not None else None}, value:{self.param_sample.value if self.param_sample is not None else None}, depth:{self.depth},  children:{len(self._children)}, is_terminal:{self.is_terminal}'

    def __str__(self, level=0):
        indent = "\t" * level

        str = f'{indent}{self.info()}\n'
        for child in self.children:
            str += child.__str__(level + 1)
        return str


class MCTree(object):
    def __init__(self, space_fn, policy, max_node_space):
        self.space_fn = space_fn
        self.policy = policy
        self.max_node_space = max_node_space
        self.root = MCNode(generate_id(), 'ROOT', param_sample=None, tree=self)
        self._current_node = self.root

    @property
    def current_node(self):
        return self._current_node

    def selection_and_expansion(self):
        node = self.root
        while not node.is_terminal:
            if node.is_leaf:
                space_sample, child = self.expansion(node)
                if child != node:
                    return space_sample, child
            else:
                # print(f'Tree selection:{node.info()}')
                node = self.policy.selection(node)
                if node.visits <= 0:
                    break

        space_sample = self.node_to_space(node)
        return space_sample, node

    def path_to_node(self, node):
        nodes = []
        while node.parent is not None:
            nodes.insert(0, node)
            node = node.parent
        return nodes

    def node_to_space(self, node):
        space_sample = self.space_fn()
        nodes = self.path_to_node(node)
        i = 0
        for hp in space_sample.params_iterator:
            if i >= len(nodes):
                break
            hp.assign(nodes[i].param_sample.value)
            i += 1
        return space_sample

    def expansion(self, node):
        # print(f'Tree expansion:{node.info()}')
        space_sample = self.space_fn()
        nodes = self.path_to_node(node)
        i = 0
        for hp in space_sample.params_iterator:
            if i < len(nodes):
                hp.assign(nodes[i].param_sample.value)
            else:
                node.expansion(hp, self.max_node_space)
                child = node.random_sample()
                hp.assign(child.param_sample.value)
                return space_sample, child
            i += 1
        node.set_terminal()
        return space_sample, node

    def simulation(self, node):
        raise NotImplementedError

    def back_propagation(self, node, reward, is_simulation=False):
        while node is not None:
            node.reward, node.visits = self.policy.back_propagation(node, reward, is_simulation)
            node = node.parent

    def roll_out(self, space_sample, node):
        # print(f'Tree roll out:{node.info()}')
        terminal = True
        for hp in space_sample.params_iterator:
            terminal = False
            hp.random_sample()
        if terminal:
            node.set_terminal()
        return space_sample


class BasePolicy(object):
    def selection(self, node):
        raise NotImplementedError

    def back_propagation(self, node, reward):
        raise NotImplementedError


class UCT(BasePolicy):
    def __init__(self, exploration_bonus=0.6):
        self.exploration_bonus = exploration_bonus

    def selection(self, node):
        node_log_nt = 0 if node.visits <= 0 else math.log10(node.visits)
        scores = [(child.reward + self.exploration_bonus * math.sqrt(
            node_log_nt / child.visits)) if child.visits > 0 else np.inf for child
                  in
                  node.children]
        details = [(child.param_sample.value, np.sum(child.rewards), child.reward, len(child.rewards)) for child in
                   node.children]
        index = np.argmax(scores)
        selected = node.children[index]
        # print(f'UCT selection: scores:{scores}, index:{index}, selected:{selected.info()}')
        # print(f'Detials: {details}')
        # print('*******************************************************************************')

        return selected

    def back_propagation(self, node, reward, is_simulation=False):
        if is_simulation:
            node.simulation_rewards.append(reward)
            # return np.average(node.rewards + [reward]), len(node.rewards) + 1
        else:
            node.rewards.append(reward)
            # return np.average(node.rewards), len(node.rewards)
        avg_reward = np.average(node.rewards) if len(node.rewards) > 0 else 0.0
        if len(node.simulation_rewards) > 0:
            avg_sim_reward = np.average(node.simulation_rewards)
            new_reward = (avg_reward + avg_sim_reward) / 2
            visits = len(node.simulation_rewards)
        else:
            new_reward = avg_reward
            visits = len(node.rewards)
        return new_reward, visits



# -*- coding:utf-8 -*-
"""

"""
import datetime
import os
import pickle
import shutil
from collections import OrderedDict

import pandas as pd

from hypernets.utils.common import isnotebook, to_repr
from ..core.searcher import OptimizeDirection


class Trial():
    def __init__(self, space_sample, trial_no, reward, elapsed, model_file=None, succeeded=True):
        self.space_sample = space_sample
        self.trial_no = trial_no
        self.reward = reward
        self.elapsed = elapsed
        self.model_file = model_file
        self.succeeded = succeeded

        self.memo = {}
        self.iteration_scores = {}

    def __repr__(self):
        return to_repr(self)

    def _repr_html_(self):
        html = f'<div><h>Trial:</h>'
        html += '''<table border="1" class="dataframe">
<thead>
<tr style="text-align: right;">
  <th>key</th>
  <th>value</th>
</tr>
</thead>
<tbody>'''
        html += f'''<tr>
  <td>Trial No.</td>
  <td>{self.trial_no}</td>
</tr>
<tr>
  <td>Reward</td>
  <td>{self.reward}</td>
</tr>
<tr>
  <td>Elapsed</td>
  <td>{self.elapsed}</td>
</tr>
<tr>
  <td>space.signature</td>
  <td>{self.space_sample.signature}</td>
</tr>
<tr>
  <td>space.vectors</td>
  <td>{self.space_sample.vectors}</td>
</tr>'''
        params = self.space_sample.get_assigned_params()
        for i, hp in enumerate(params):
            html += f'''<tr>
  <td>{i}-{hp.alias}</td>
  <td>{hp.value}</td>
</tr>
<tr>'''

        html += '''  </tbody>
</table>
</div>'''
        return html

    def __getstate__(self):
        try:
            state = super().__getstate__()
        except AttributeError:
            state = self.__dict__

        state = {k: v for k, v in state.items() if k != 'memo'}
        return state

    def to_df(self, include_params=False):
        out = OrderedDict(trial_no=self.trial_no, succeeded=self.succeeded, reward=self.reward, elapsed=self.elapsed)
        if include_params:
            for p in self.space_sample.get_assigned_params():
                out[p.alias] = p.value
        return pd.DataFrame({k: [v] for k, v in out.items()})


class TrialHistory():
    def __init__(self, optimize_direction):
        self.trials = []
        self.optimize_direction = optimize_direction

    def append(self, trial):
        old_best = self.get_best()
        self.trials.append(trial)
        new_best = self.get_best()
        improved = old_best != new_best
        return improved

    def is_existed(self, space_sample):
        return space_sample.vectors in [t.space_sample.vectors for t in self.trials]

    def get_trial(self, space_sample):
        all_vectors = [t.space_sample.vectors for t in self.trials]
        index = all_vectors.index(space_sample.vectors)
        if index >= 0:
            return self.trials[index]
        else:
            return None

    def get_best(self):
        top1 = self.get_top(1)
        if len(top1) <= 0:
            return None
        else:
            return top1[0]

    def get_worst(self):
        topn = self.get_top()
        return topn[-1] if len(topn) > 0 else None

    def get_top(self, n=None):
        assert n is None or isinstance(n, int)

        valid_trials = [t for t in self.trials if t.succeeded]
        if len(valid_trials) <= 0:
            return []
        sorted_trials = sorted(valid_trials, key=lambda t: t.reward,
                               reverse=self.optimize_direction in ['max', OptimizeDirection.Maximize])

        if isinstance(n, int) and n < len(sorted_trials):
            sorted_trials = sorted_trials[:n]

        return sorted_trials

    def get_space_signatures(self):
        signatures = set()
        for s in [t.space_sample for t in self.trials]:
            signatures.add(s.signature)
        return signatures

    def diff(self, trials):
        signatures = set()
        for s in [t.space_sample for t in trials]:
            signatures.add(s.signature)

        diffs = {}
        for sign in signatures:
            ts = [t for t in trials if t.space_sample.signature == sign]
            pv_dict = {}
            for t in ts:
                for p in t.space_sample.get_assigned_params():
                    k = p.alias
                    v = str(p.value)
                    if pv_dict.get(k) is None:
                        pv_dict[k] = {}
                    if pv_dict[k].get(v) is None:
                        pv_dict[k][v] = [t.reward]
                    else:
                        pv_dict[k][v].append(t.reward)
            diffs[sign] = pv_dict

        return diffs

    def get_trajectories(self):
        times, best_rewards, rewards = [0.0], [0.0], [0.0]
        his = sorted(self.trials, key=lambda t: t.trial_no)
        best_trial_no = 0
        best_elapsed = 0
        for t in his:
            rewards.append(t.reward)
            times.append(t.elapsed + times[-1])

            if t.reward > best_rewards[-1]:
                best_rewards.append(t.reward)
                best_trial_no = t.trial_no
                best_elapsed = times[-1]
            else:
                best_rewards.append(best_rewards[-1])

        return times, best_rewards, rewards, best_trial_no, best_elapsed

    def save(self, filepath):
        with open(filepath, 'w') as output:
            output.write(f'{self.optimize_direction}\r\n')
            for trial in self.trials:
                data = f'{trial.trial_no}|{trial.space_sample.vectors}|{trial.reward}|{trial.elapsed}' + \
                       f'|{trial.model_file if trial.model_file else ""}\r\n'
                output.write(data)

    @staticmethod
    def load_history(space_fn, filepath):
        with open(filepath, 'r') as input:
            line = input.readline()
            history = TrialHistory(line.strip())
            while line is not None and line != '':
                line = input.readline()
                if line.strip() == '':
                    continue
                fields = line.split('|')
                assert len(fields) >= 4, f'Trial format is not correct. \r\nline:[{line}]'
                sample = space_fn()
                vector = [float(n) if n.__contains__('.') else int(n) for n in
                          fields[1].replace('[', '').replace(']', '').split(',')]
                sample.assign_by_vectors(vector)
                if len(fields) > 4:
                    model_file = fields[4]
                else:
                    model_file = None
                trial = Trial(space_sample=sample, trial_no=int(fields[0]), reward=float(fields[2]),
                              elapsed=float(fields[3]), model_file=model_file)
                history.append(trial)
            return history

    def __repr__(self):
        out = OrderedDict(direction=self.optimize_direction)
        if len(self.trials) > 0:
            tops = self.get_top()
            out['size'] = len(self.trials)
            out['succeeded'] = len(tops)
            if len(tops) > 0:
                out['best_reward'] = tops[0].reward
                out['worst_reward'] = tops[-1].reward

        repr_ = ', '.join('%s=%r' % (k, v) for k, v in out.items())
        return f'{type(self).__name__}({repr_})'

    def to_df(self, include_params=False):
        if len(self.trials) > 0:
            df = pd.concat([t.to_df(include_params) for t in self.trials], axis=0)
            df.reset_index(drop=True, inplace=True)
        else:
            df = pd.DataFrame()

        return df

    def plot_hyperparams(self, destination='notebook', output='hyperparams.html'):
        """Plot hyperparams in a parallel line chart

        Args:
            destination: one of notebook, html
            output: the html file path

        Returns:

        """

        try:
            import pyecharts
        except Exception as e:
            raise Exception("You may not install 'pyecharts',"
                            "please refer to https://github.com/pyecharts/pyecharts and install it.")

        import pyecharts.options as opts
        from pyecharts.charts import Parallel

        if destination == 'notebook' and not isnotebook():
            raise Exception("You may not running in a notebook,"
                            " try to set 'destination' to 'html' or run it in notebook ")

        if self.trials is None or len(self.trials) < 1:
            raise Exception("Trials is empty ")

        REWARD_METRIC_COL = 'Reward metric'

        def get_space_params(trial):
            space = trial.space_sample
            params_dict = {}
            for hyper_param in space.get_all_params():
                references = list(hyper_param.references)
                if len(references) > 0:
                    param_name = hyper_param.alias[len(list(hyper_param.references)[0].name) + 1:]
                    param_value = hyper_param.value

                    if isinstance(param_value, int) or isinstance(param_value, float):
                        if not isinstance(param_value, bool):
                            params_dict[param_name] = param_value
                params_dict[REWARD_METRIC_COL] = trial.reward

            return params_dict

        def make_dims(df_params):
            parallel_axis = []
            for i, col in enumerate(df_params.columns):
                if df_params.dtypes[col].kind == 'O':
                    parallel_axis.append({
                        "dim": i,
                        "name": col,
                        "type": "category",
                        "data": df_params[col].unique().tolist(),
                    })
                else:
                    parallel_axis.append({'dim': i, 'name': col})
            return parallel_axis

        trials_params = [get_space_params(trial) for trial in self.trials]

        param_names = list(set([v for ps in trials_params for v in ps.keys()]))
        param_names.remove(REWARD_METRIC_COL)
        param_names.insert(len(param_names), REWARD_METRIC_COL)

        trial_params_values = []
        for t in trials_params:
            param_values = [t.get(n) for n in param_names]
            trial_params_values.append(param_values)

        df_train_params = pd.DataFrame(data=trial_params_values, columns=param_names)

        # remove if all is None
        df_train_params.dropna(axis=1, how='all', inplace=True)

        parallel_axis = make_dims(df_train_params)
        chart = \
            Parallel(init_opts=opts.InitOpts(width="%dpx" % (len(param_names) * 100), height="400px")) \
                .add_schema(schema=parallel_axis).add(series_name="",
                                                      data=df_train_params.values.tolist(),
                                                      linestyle_opts=opts.LineStyleOpts(width=1, opacity=0.5),

                                                      ).set_global_opts(
                visualmap_opts=[
                    opts.VisualMapOpts(
                        type_="color",
                        is_calculable=True,
                        precision=2,
                        # dimension=0,
                        pos_left="-10",
                        pos_bottom="30",
                        max_=df_train_params[REWARD_METRIC_COL].max().tolist(),
                        min_=df_train_params[REWARD_METRIC_COL].min().tolist()
                    )
                ]
            )

        if destination == 'notebook':
            return chart.render_notebook()
        else:
            return chart.render(output)


class TrialStore(object):
    def __init__(self):
        self.reset()
        self.load()

    def put(self, dataset_id, trial):
        self.put_to_cache(dataset_id, trial)
        self._put(dataset_id, trial)

    def get_from_cache(self, dataset_id, space_sample):
        if self._cache.get(dataset_id) is None:
            self._cache[dataset_id] = {}
        dataset = self._cache[dataset_id]
        if dataset.get(space_sample.signature) is None:
            dataset[space_sample.signature] = {}
        trial = dataset[space_sample.signature].get(self.sample2key(space_sample))
        return trial

    def put_to_cache(self, dataset_id, trial):
        if self._cache.get(dataset_id) is None:
            self._cache[dataset_id] = {}
        dataset = self._cache[dataset_id]
        if dataset.get(trial.space_sample.signature) is None:
            dataset[trial.space_sample.signature] = {}
        dataset[trial.space_sample.signature][self.sample2key(trial.space_sample)] = trial

    def get(self, dataset_id, space_sample):
        trial = self.get_from_cache(dataset_id, space_sample)
        if trial is None:
            trial = self._get(dataset_id, space_sample)
            if trial is not None:
                self.put_to_cache(dataset_id, trial)
        return trial

    def _get(self, dataset_id, space_sample):
        raise NotImplementedError

    def _put(self, dataset_id, trial):
        raise NotImplementedError

    def load(self):
        raise NotImplementedError

    def get_all(self, dataset_id, space_signature):
        raise NotImplementedError

    def reset(self):
        raise NotImplementedError

    def persist(self):
        raise NotImplementedError

    def sample2key(self, space_sample):
        key = ','.join([str(f) for f in space_sample.vectors])
        return key

    def check_trial(self, trial):
        pass


class DiskTrialStore(TrialStore):
    def __init__(self, home_dir=None):
        self.home_dir = self.prepare_home_dir(home_dir)
        TrialStore.__init__(self)

    def prepare_home_dir(self, home_dir):
        if home_dir is None:
            home_dir = 'trial_store'
        if home_dir[-1] == '/':
            home_dir = home_dir[:-1]
        home_dir = os.path.expanduser(home_dir)
        if not os.path.exists(home_dir):
            os.makedirs(home_dir)
        return home_dir

    def _prepare_output_dir(self, log_dir, searcher):
        if log_dir is None:
            log_dir = 'log'
        if log_dir[-1] == '/':
            log_dir = log_dir[:-1]

        running_dir = f'exp_{searcher.__class__.__name__}_{datetime.datetime.now().__format__("%m%d-%H%M%S")}'
        output_path = os.path.expanduser(f'{log_dir}/{running_dir}/')
        if not os.path.exists(output_path):
            os.makedirs(output_path)
        return output_path

    def load(self):
        pass

    def clear_history(self):
        shutil.rmtree(self.home_dir)
        self.prepare_home_dir(self.home_dir)
        self.reset()

    def reset(self):
        self._cache = {}

    def _get(self, dataset_id, space_sample):
        path = self.get_trial_path(dataset_id, space_sample)
        trial = self._load_trial(path)
        if trial is not None:
            trial.space_sample = space_sample
        return trial

    def _load_trial(self, path):
        if not os.path.exists(path):
            return None
        else:
            with open(path, 'rb') as f:
                trial = pickle.load(f)
                self.check_trial(trial)
            return trial

    def _put(self, dataset_id, trial):
        path = self.get_trial_path(dataset_id, trial.space_sample)
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        with open(path, 'wb') as f:
            temp = Trial(space_sample=None,
                         trial_no=trial.trial_no,
                         reward=trial.reward,
                         elapsed=trial.elapsed,
                         model_file=trial.model_file)
            temp.space_sample_vectors = trial.space_sample.vectors
            pickle.dump(temp, f)

    def persist(self):
        pass

    def get_trial_path(self, dataset_id, space_sample):
        path = f'{self.home_dir}/{dataset_id}/{space_sample.signature}/{self.sample2key(space_sample)}.pkl'
        return path

    def get_all(self, dataset_id, space_signature):
        path = f'{self.home_dir}/{dataset_id}/{space_signature}'
        trials = []
        if not os.path.exists(path):
            return trials
        files = os.listdir(path)
        for f in files:
            if f.endswith('.pkl'):
                f = path + '/' + f
                trial = self._load_trial(f)
                trials.append(trial)
        return trials


default_trial_store = None


def get_default_trial_store():
    return default_trial_store


def set_default_trial_store(trial_store):
    global default_trial_store
    default_trial_store = trial_store



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
from .base_ensemble import BaseEnsemble
from sklearn.linear_model import LogisticRegression, LinearRegression

import numpy as np


class StackingEnsemble(BaseEnsemble):
    def __init__(self, task, estimators, need_fit=False, n_folds=5, method='soft', meta_model=None, fit_kwargs=None):
        super(StackingEnsemble, self).__init__(task, estimators, need_fit, n_folds, method)
        if meta_model is None:
            if task == 'regression':
                self.meta_model = LinearRegression()
            else:
                self.meta_model = LogisticRegression()
        else:
            self.meta_model = meta_model
        self.fit_kwargs = fit_kwargs if fit_kwargs is not None else {}

    def fit_predictions(self, predictions, y_true):
        X = self.__predictions2X(predictions)
        self.meta_model.fit(X, y_true, **self.fit_kwargs)

    def __predictions2X(self, predictions):
        X = predictions
        if len(X.shape) == 3:
            if self.task == 'binary':
                X = X[:, :, -1]
            elif self.task == 'multiclass':
                X = np.argmax(X, axis=2)
            else:
                raise ValueError(
                    f"The shape of `predictions` and the `task` don't match. shape:{predictions.shape}, task:{self.task}")
        return X

    def predictions2predict(self, predictions):
        assert self.meta_model is not None
        X = self.__predictions2X(predictions)
        pred = self.meta_model.predict(X)
        if self.task == 'binary':
            pred = np.clip(pred, 0, 1)
        return pred

    def predictions2predict_proba(self, predictions):
        assert self.meta_model is not None
        X = self.__predictions2X(predictions)
        if hasattr(self.meta_model, 'predict_proba'):
            pred = self.meta_model.predict_proba(X)
        else:
            pred = self.meta_model.predict(X)

        if self.task == 'binary':
            pred = np.clip(pred, 0, 1)
        return pred



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from . import BaseEstimator
import h2o
from h2o.automl import H2OAutoML


class H2OEstimator(BaseEstimator):
    def __init__(self, task, **kwargs):
        super(H2OEstimator, self).__init__(task)
        self.name = 'H2O AutoML'
        self.kwargs = kwargs
        self.estimator = None

    def train(self, X, y, X_test):
        h2o.init()
        target = '__tabular_toolbox_target__'
        X.insert(0, target, y)
        train = h2o.H2OFrame(X)
        x_cols = train.columns
        x_cols.remove(target)
        train[target] = train[target].asfactor()
        self.esitmator = H2OAutoML(max_models=20, seed=1)
        self.esitmator.train(x=x_cols, y=target, training_frame=train)

    def predict_proba(self, X):
        x = h2o.H2OFrame(X)
        preds = self.esitmator.predict(x)
        preds = preds[1:].as_data_frame().values
        return preds

    def predict(self, X):
        proba = self.predict_proba(X)
        return self.proba2predict(proba)



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
import copy
import io

import pandas as pd

from hypernets.utils import hash_dataframe

csv_str = '''x1_int_nanchar,x2_all_nan,x3_const_str,x4_const_int,x5_dup_1,x6_dup_2,x7_dup_f1,x8_dup_f2,x9_f,x10,y
1.0,,const,5,dup,dup,0.1,0.1,1.23,\\N,1
2.2,,const,5,dupa,dupa,0.111,0.111,4.4,\\N,1
\\N,,const,5,dupb,dupb,0.12323,0.12323,1.233,\\N,1
4.,,const,5,dupc,dupc,0.14334,0.14334,4534434.2,\\N,0
5,,const,5,dupd,dupd,0.144,0.144,2302.2,\\N,0
6,,const,5,dupe,dupe,0.155,0.155,34334.1,\\N,\\N
'''


class Test_DataCleaner():
    def test_basic(self):
        df1 = pd.read_csv(io.StringIO(csv_str))
        hash1 = hash_dataframe(df1)

        df2 = pd.read_csv(io.StringIO(csv_str))
        hash2 = hash_dataframe(df2)
        assert hash1 == hash2

        df3 = df1.head(5)
        hash3 = hash_dataframe(df3)
        assert hash1 != hash3

        df4 = pd.concat([df1, df1.head(1)], axis=0)
        hash4 = hash_dataframe(df4)
        assert hash1 != hash4

        df5 = copy.deepcopy(df1)
        df5['x1_int_nanchar'] = ['1.0', '2.2', '\\N', '4.', '5', '6']
        hash5 = hash_dataframe(df5)
        assert hash1 == hash5

        df6 = copy.deepcopy(df1)
        df6['x1_int_nanchar'] = ['2.0', '2.2', '\\N', '4.', '5', '6']
        hash6 = hash_dataframe(df6)
        assert hash1 != hash6

    # TODO @lxf add unit tests for Dask.DataFrame



# -*- coding:utf-8 -*-

import time

from hypernets.dispatchers.cluster.grpc.search_driver_service import get_or_serve
from hypernets.core.callbacks import EarlyStoppingError
from hypernets.core.dispatcher import Dispatcher
from hypernets.core.trial import Trial
from hypernets.dispatchers.cfg import DispatchCfg as c
from hypernets.utils import logging

logger = logging.get_logger(__name__)

_search_counter = 0


class DriverDispatcher(Dispatcher):
    def __init__(self, address, work_dir):
        super(DriverDispatcher, self).__init__()

        self.address = address
        self.work_dir = work_dir
        self.spaces_dir = f'{work_dir}/spaces'
        self.models_dir = f'{work_dir}/models'

    def dispatch(self, hyper_model, X, y, X_eval, y_eval, cv, num_folds, max_trials, dataset_id, trial_store,
                 **fit_kwargs):
        def on_next_space(item):
            for cb in hyper_model.callbacks:
                # cb.on_build_estimator(hyper_model, space_sample, estimator, trial_no)
                cb.on_trial_begin(hyper_model, item.space_sample, item.trial_no)

        def on_report_space(item):
            if item.success:
                elapsed = item.report_at - item.start_at
                trial = Trial(item.space_sample, item.trial_no, item.reward, elapsed)
                # print(f'trial result:{trial}')

                improved = hyper_model.history.append(trial)
                if improved and logger.is_info_enabled():
                    logger.info(
                        f'>>>improved: reward={item.reward}, trial_no={item.trial_no}, space_id={item.space_id}')
                hyper_model.searcher.update_result(item.space_sample, item.reward)

                if trial_store is not None:
                    trial_store.put(dataset_id, trial)

                for cb in hyper_model.callbacks:
                    cb.on_trial_end(hyper_model, item.space_sample, item.trial_no, item.reward, improved, elapsed)
            else:
                for cb in hyper_model.callbacks:
                    cb.on_trial_error(hyper_model, space_sample, item.trial_no)

        def on_summary():
            t = hyper_model.get_best_trial()
            if t:
                detail = f'reward={t.reward}, trial_no={t.trial_no}, space_id={t.space_sample.space_id}'
                return f'best: {detail}'
            else:
                return None

        def do_clean():
            # shutdown grpc server
            search_service.status_thread.stop()
            search_service.status_thread.report_summary()
            # server.stop(grace=1.0)

        if 'search_id' in fit_kwargs:
            search_id = fit_kwargs.pop('search_id')
        else:
            global _search_counter
            _search_counter += 1
            search_id = 'search-%02d' % _search_counter

        if logger.is_info_enabled():
            logger.info(f'start driver server at {self.address}')
        server, search_service = get_or_serve(self.address,
                                              search_id,
                                              self.spaces_dir,
                                              self.models_dir,
                                              on_next=on_next_space,
                                              on_report=on_report_space,
                                              on_summary=on_summary)

        search_start_at = time.time()

        trial_no = 1
        retry_counter = 0
        queue_size = c.cluster_search_queue

        while trial_no <= max_trials:
            space_sample = hyper_model.searcher.sample()
            if hyper_model.history.is_existed(space_sample):
                if retry_counter >= 1000:
                    if logger.is_info_enabled():
                        logger.info(f'Unable to take valid sample and exceed the retry limit 1000.')
                    break
                trial = hyper_model.history.get_trial(space_sample)
                for callback in hyper_model.callbacks:
                    callback.on_skip_trial(hyper_model, space_sample, trial_no, 'trial_exsited', trial.reward, False,
                                           trial.elapsed)
                retry_counter += 1
                continue

            try:
                if trial_store is not None:
                    trial = trial_store.get(dataset_id, space_sample)
                    if trial is not None:
                        reward = trial.reward
                        elapsed = trial.elapsed
                        trial = Trial(space_sample, trial_no, reward, elapsed)
                        improved = hyper_model.history.append(trial)
                        hyper_model.searcher.update_result(space_sample, reward)
                        for callback in hyper_model.callbacks:
                            callback.on_skip_trial(hyper_model, space_sample, trial_no, 'hit_trial_store', reward,
                                                   improved,
                                                   elapsed)
                        trial_no += 1
                        continue

                search_service.add(trial_no, space_sample)

                # wait for queued trial
                while search_service.queue_size() >= queue_size:
                    time.sleep(0.1)
            except EarlyStoppingError:
                break
                # TODO: early stopping
            except KeyboardInterrupt:
                do_clean()
                return trial_no
            except Exception as e:
                if logger.is_warning_enabled():
                    import sys
                    import traceback
                    msg = f'{e.__class__.__name__}: {e}'
                    logger.warning(f'{">" * 20} Trial failed! {"<" * 20}')
                    logger.warning(msg + '\n' + traceback.format_exc())
                    logger.warning('*' * 50)
            finally:
                trial_no += 1
                retry_counter = 0
        if logger.is_info_enabled():
            logger.info("-" * 20 + 'no more space to search, waiting trials ...')
        try:
            while search_service.running_size() > 0:
                # if logger.is_info_enabled():
                #    logger.info(f"wait ... {search_service.running_size()} samples found.")
                time.sleep(0.1)
        except KeyboardInterrupt:
            return trial_no
        finally:
            do_clean()

        if logger.is_info_enabled():
            logger.info('-' * 20 + ' all trials done ' + '-' * 20)

        return trial_no



import re
from collections import defaultdict
from functools import partial

import dask
import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd
from dask_ml import preprocessing as dm_pre, decomposition as dm_dec
from sklearn import preprocessing as sk_pre
from sklearn.base import BaseEstimator, TransformerMixin

from hypernets.tabular import sklearn_ex as skex
from hypernets.utils import logging, const

logger = logging.get_logger(__name__)


#
# class MultiLabelEncoder(BaseEstimator, TransformerMixin):
#     def __init__(self):
#         self.encoders = {}
#
#     def fit(self, X, y=None):
#         assert len(X.shape) == 2
#
#         if isinstance(X, (pd.DataFrame, dd.DataFrame)):
#             return self._fit_df(X, y)
#         elif isinstance(X, (np.ndarray, da.Array)):
#             return self._fit_array(X, y)
#         else:
#             raise Exception(f'Unsupported type "{type(X)}"')
#
#     def _fit_df(self, X, y=None):
#         return self._fit_array(X.values, y.values if y else None)
#
#     def _fit_array(self, X, y=None):
#         n_features = X.shape[1]
#         for n in range(n_features):
#             le = dm_pre.LabelEncoder()
#             le.fit(X[:, n])
#             self.encoders[n] = le
#         return self
#
#     def transform(self, X):
#         assert len(X.shape) == 2
#
#         if isinstance(X, (dd.DataFrame, pd.DataFrame)):
#             return self._transform_dask_df(X)
#         elif isinstance(X, (da.Array, np.ndarray)):
#             return self._transform_dask_array(X)
#         else:
#             raise Exception(f'Unsupported type "{type(X)}"')
#
#     def _transform_dask_df(self, X):
#         data = self._transform_dask_array(X.values)
#
#         if isinstance(X, dd.DataFrame):
#             result = dd.from_dask_array(data, columns=X.columns)
#         else:
#             result = pd.DataFrame(data, columns=X.columns)
#         return result
#
#     def _transform_dask_array(self, X):
#         n_features = X.shape[1]
#         assert n_features == len(self.encoders.items())
#
#         data = []
#         for n in range(n_features):
#             data.append(self.encoders[n].transform(X[:, n]))
#
#         if isinstance(X, da.Array):
#             result = da.stack(data, axis=-1, allow_unknown_chunksizes=True)
#         else:
#             result = np.stack(data, axis=-1)
#
#         return result
#
#     # def fit_transform(self, X, y=None):
#     #     return self.fit(X, y).transform(X)


class SafeOneHotEncoder(dm_pre.OneHotEncoder):
    def fit(self, X, y=None):
        if isinstance(X, (dd.DataFrame, pd.DataFrame)) and self.categories == "auto" \
                and any(d.name in {'object', 'bool'} for d in X.dtypes):
            a = []
            if isinstance(X, dd.DataFrame):
                for i in range(len(X.columns)):
                    Xi = X.iloc[:, i]
                    if Xi.dtype.name in {'object', 'bool'}:
                        Xi = Xi.astype('category').cat.as_known()
                    a.append(Xi)
                X = dd.concat(a, axis=1, ignore_unknown_divisions=True)
            else:
                for i in range(len(X.columns)):
                    Xi = X.iloc[:, i]
                    if Xi.dtype.name in {'object', 'bool'}:
                        Xi = Xi.astype('category')
                    a.append(Xi)
                X = pd.concat(a, axis=1)

        return super(SafeOneHotEncoder, self).fit(X, y)

    def get_feature_names(self, input_features=None):
        """
        Override this method to remove non-alphanumeric chars
        """
        # if not hasattr(self, 'drop_idx_'):
        #     setattr(self, 'drop_idx_', None)
        # return super(SafeOneHotEncoder, self).get_feature_names(input_features)

        # check_is_fitted(self)
        cats = self.categories_
        if input_features is None:
            input_features = ['x%d' % i for i in range(len(cats))]
        elif len(input_features) != len(self.categories_):
            raise ValueError(
                "input_features should have length equal to number of "
                "features ({}), got {}".format(len(self.categories_),
                                               len(input_features)))

        feature_names = []
        for i in range(len(cats)):
            names = [input_features[i] + '_' + str(idx) + '_' + re.sub('[^A-Za-z0-9_]+', '_', str(t))
                     for idx, t in enumerate(cats[i])]
            # if self.drop_idx_ is not None and self.drop_idx_[i] is not None:
            #     names.pop(self.drop_idx_[i])
            feature_names.extend(names)

        return np.array(feature_names, dtype=object)


class TruncatedSVD(dm_dec.TruncatedSVD):
    def fit_transform(self, X, y=None):
        X_orignal = X
        if isinstance(X, pd.DataFrame):
            X = dd.from_pandas(X, npartitions=2)

        if isinstance(X, dd.DataFrame):
            # y = y.values.compute_chunk_sizes() if y is not None else None
            r = super(TruncatedSVD, self).fit_transform(X.values.compute_chunk_sizes(), None)
        else:
            r = super(TruncatedSVD, self).fit_transform(X, y)

        if isinstance(X_orignal, (pd.DataFrame, np.ndarray)):
            r = r.compute()
        return r

    def transform(self, X, y=None):
        if isinstance(X, dd.DataFrame):
            return super(TruncatedSVD, self).transform(X.values, y)

        return super(TruncatedSVD, self).transform(X, y)

    def inverse_transform(self, X):
        if isinstance(X, dd.DataFrame):
            return super(TruncatedSVD, self).inverse_transform(X.values)

        return super(TruncatedSVD, self).inverse_transform(X)


class MaxAbsScaler(sk_pre.MaxAbsScaler):
    __doc__ = sk_pre.MaxAbsScaler.__doc__

    def fit(self, X, y=None, ):
        from dask_ml.utils import handle_zeros_in_scale

        self._reset()
        if isinstance(X, (pd.DataFrame, np.ndarray)):
            return super().fit(X, y)

        max_abs = X.reduction(lambda x: x.abs().max(),
                              aggregate=lambda x: x.max(),
                              token=f'{self.__class__.__name__}.fit'
                              ).compute()
        scale = handle_zeros_in_scale(max_abs)

        setattr(self, 'max_abs_', max_abs)
        setattr(self, 'scale_', scale)
        setattr(self, 'n_samples_seen_', 0)

        self.n_features_in_ = X.shape[1]
        return self

    def partial_fit(self, X, y=None, ):
        raise NotImplementedError()

    def transform(self, X, y=None, copy=None, ):
        if isinstance(X, (pd.DataFrame, np.ndarray)):
            return super().transform(X)

        # Workaround for https://github.com/dask/dask/issues/2840
        if isinstance(X, dd.DataFrame):
            X = X.div(self.scale_)
        else:
            X = X / self.scale_
        return X

    def inverse_transform(self, X, y=None, copy=None, ):
        if not hasattr(self, "scale_"):
            raise Exception(
                "This %(name)s instance is not fitted yet. "
                "Call 'fit' with appropriate arguments before "
                "using this method."
            )

        if isinstance(X, (pd.DataFrame, np.ndarray)):
            return super().inverse_transform(X)

        if copy:
            X = X.copy()
        if isinstance(X, dd.DataFrame):
            X = X.mul(self.scale_)
        else:
            X = X * self.scale_

        return X


def _safe_ordinal_encoder(categories, dtype, pdf):
    assert isinstance(pdf, pd.DataFrame)

    def encode_column(x, c):
        return c[x]

    mappings = {}
    for col, cat in categories.items():
        unseen = len(cat)
        m = defaultdict(dtype)
        for k, v in zip(cat, range(unseen)):
            m[k] = dtype(v + 1)
        mappings[col] = m

    pdf = pdf.copy()
    vf = np.vectorize(encode_column, excluded='c', otypes=[dtype])
    for col, m in mappings.items():
        r = vf(pdf[col].values, m)
        if r.dtype != dtype:
            # print(r.dtype, 'astype', dtype)
            r = r.astype(dtype)
        pdf[col] = r
    return pdf


def _safe_ordinal_decoder(categories, dtypes, pdf):
    assert isinstance(pdf, pd.DataFrame)

    def decode_column(x, col):
        cat = categories[col]
        xi = int(x)
        unseen = cat.shape[0]  # len(cat)
        if unseen >= xi >= 1:
            return cat[xi - 1]
        else:
            dtype = dtypes[col]
            if dtype in (np.float32, np.float64, float):
                return np.nan
            elif dtype in (np.int32, np.int64, np.int, np.uint32, np.uint64, np.uint):
                return -1
            else:
                return None

    pdf = pdf.copy()
    for col in categories.keys():
        vf = np.vectorize(decode_column, excluded='col', otypes=[dtypes[col]])
        pdf[col] = vf(pdf[col].values, col)
    return pdf


class SafeOrdinalEncoder(BaseEstimator, TransformerMixin):
    __doc__ = r'Adapted from dask_ml OrdinalEncoder\n' + dm_pre.OrdinalEncoder.__doc__

    def __init__(self, columns=None, dtype=np.float64):
        self.columns = columns
        self.dtype = dtype

        # fitted
        self.columns_ = None
        self.dtypes_ = None
        self.categorical_columns_ = None
        self.non_categorical_columns_ = None
        self.categories_ = None

    def fit(self, X, y=None):
        self.columns_ = X.columns.to_list()
        self.dtypes_ = {c: X[c].dtype for c in X.columns}

        if self.columns is None:
            columns = X.select_dtypes(include=["category", 'object', 'bool']).columns.to_list()
        else:
            columns = self.columns

        X = X.categorize(columns=columns)

        self.categorical_columns_ = columns
        self.non_categorical_columns_ = X.columns.drop(self.categorical_columns_)
        self.categories_ = {c: X[c].cat.categories.sort_values() for c in columns}

        return self

    def transform(self, X, y=None):
        if X.columns.to_list() != self.columns_:
            raise ValueError(
                "Columns of 'X' do not match the training "
                "columns. Got {!r}, expected {!r}".format(X.columns.to_list(), self.columns)
            )

        encoder = self.make_encoder(self.categories_, self.dtype)
        if isinstance(X, pd.DataFrame):
            X = encoder(X)
        elif isinstance(X, dd.DataFrame):
            X = X.map_partitions(encoder)
        else:
            raise TypeError("Unexpected type {}".format(type(X)))

        return X

    def inverse_transform(self, X, missing_value=None):
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X, columns=self.categorical_columns_)
        elif isinstance(X, da.Array):
            # later on we concat(..., axis=1), which requires
            # known divisions. Suboptimal, but I think unavoidable.
            unknown = np.isnan(X.chunks[0]).any()
            if unknown:
                lengths = da.blockwise(len, "i", X[:, 0], "i", dtype="i8").compute()
                X = X.copy()
                chunks = (tuple(lengths), X.chunks[1])
                X._chunks = chunks
            X = dd.from_dask_array(X, columns=self.categorical_columns_)

        decoder = self.make_decoder(self.categories_, self.dtypes_)

        if isinstance(X, dd.DataFrame):
            X = X.map_partitions(decoder)
        else:
            X = decoder(X)

        return X

    @staticmethod
    def make_encoder(categories, dtype):
        return partial(_safe_ordinal_encoder, categories, dtype)

    @staticmethod
    def make_decoder(categories, dtypes):
        return partial(_safe_ordinal_decoder, categories, dtypes)


# alias
MultiLabelEncoder = SafeOrdinalEncoder


class DataInterceptEncoder(BaseEstimator, TransformerMixin):

    def __init__(self, fit=False, fit_transform=False, transform=False, inverse_transform=False):
        self._intercept_fit = fit
        self._intercept_fit_transform = fit_transform
        self._intercept_transform = transform
        self._intercept_inverse_transform = inverse_transform

        super(DataInterceptEncoder, self).__init__()

    def fit(self, X, *args, **kwargs):
        if self._intercept_fit:
            self.intercept(X, *args, **kwargs)

        return self

    def fit_transform(self, X, *args, **kwargs):
        if self._intercept_fit_transform:
            X = self.intercept(X, *args, **kwargs)

        return X

    def transform(self, X, *args, **kwargs):
        if self._intercept_transform:
            X = self.intercept(X, *args, **kwargs)

        return X

    def inverse_transform(self, X, *args, **kwargs):
        if self._intercept_inverse_transform:
            X = self.intercept(X, *args, **kwargs)

        return X

    def intercept(self, X, *args, **kwargs):
        raise NotImplementedError()


class CallableAdapterEncoder(DataInterceptEncoder):
    def __init__(self, fn, **kwargs):
        assert callable(fn)

        self.fn = fn

        super(CallableAdapterEncoder, self).__init__(**kwargs)

    def intercept(self, X, *args, **kwargs):
        return self.fn(X, *args, **kwargs)


class DataCacher(DataInterceptEncoder):
    """
    persist and cache dask dataframe and array
    """

    def __init__(self, cache_dict, cache_key, remove_keys=None, **kwargs):
        assert isinstance(cache_dict, dict)

        if isinstance(remove_keys, str):
            remove_keys = set(remove_keys.split(','))

        self._cache_dict = cache_dict
        self.cache_key = cache_key
        self.remove_keys = remove_keys

        super(DataCacher, self).__init__(**kwargs)

    def intercept(self, X, *args, **kwargs):
        if self.cache_key:
            if isinstance(X, (dd.DataFrame, da.Array)):
                if logger.is_debug_enabled():
                    logger.debug(f'persist and cache {X._name} as {self.cache_key}')

                X = X.persist()

            self._cache_dict[self.cache_key] = X

        if self.remove_keys:
            for key in self.remove_keys:
                if key in self._cache_dict.keys():
                    if logger.is_debug_enabled():
                        logger.debug(f'remove cache {key}')
                    del self._cache_dict[key]

        return X

    @property
    def cache_dict(self):
        return list(self._cache_dict.keys())


class CacheCleaner(DataInterceptEncoder):

    def __init__(self, cache_dict, **kwargs):
        assert isinstance(cache_dict, dict)

        self._cache_dict = cache_dict

        super(CacheCleaner, self).__init__(**kwargs)

    def intercept(self, X, *args, **kwargs):
        if logger.is_debug_enabled():
            logger.debug(f'clean cache with {list(self._cache_dict.keys())}')
        self._cache_dict.clear()

        return X

    @property
    def cache_dict(self):
        return list(self._cache_dict.keys())

    # # override this to remove 'cache_dict' from estimator __expr__
    # @classmethod
    # def _get_param_names(cls):
    #     params = super()._get_param_names()
    #     return [p for p in params if p != 'cache_dict']


class DataFrameWrapper(skex.DataFrameWrapper):

    def transform(self, X):
        transformed = self.transformer.transform(X)

        if isinstance(transformed, da.Array):
            transformed = dd.from_dask_array(transformed, columns=self.columns)
        elif isinstance(transformed, np.ndarray):
            transformed = pd.DataFrame(transformed, columns=self.columns)
        else:
            transformed.columns = self.columns

        return transformed


class AdaptedTransformer(BaseEstimator):
    """
    Adapt sklearn style transformer to support dask data objects.
    Note: adapted transformer uses locale resource only.
    """

    def __init__(self, cls, attributes, *args, **kwargs):
        super(AdaptedTransformer, self).__init__()

        assert isinstance(cls, type)
        assert attributes is None or isinstance(attributes, (tuple, list, set))

        self.adapted_transformer_ = cls(*args, **kwargs)
        self.adapted_attributes_ = set(attributes) if attributes is not None else None

    def fit(self, X, y=None, **kwargs):
        if y is None:
            self._adapt(self.adapted_transformer_.fit, X, **kwargs)
        else:
            self._adapt(self.adapted_transformer_.fit, X, y, **kwargs)
        return self

    def transform(self, X, *args, **kwargs):
        return self._adapt(self.adapted_transformer_.transform, X, *args, **kwargs)

    def inverse_transform(self, X, *args, **kwargs):
        return self._adapt(self.adapted_transformer_.inverse_transform, X, *args, **kwargs)

    def fit_transform(self, X, y=None, **kwargs):
        transformer = self.adapted_transformer_

        if hasattr(transformer, 'fit_transform'):
            if y is None:
                return self._adapt(transformer.fit_transform, X, **kwargs)
            else:
                return self._adapt(transformer.fit_transform, X, y, **kwargs)
        else:
            self.fit(X, y, **kwargs)
            return self._adapt(transformer.transform, X, **kwargs)

    @staticmethod
    def _adapt(fn, X, *args, **kwargs):
        is_dask_frame_X = isinstance(X, (dd.DataFrame, dd.Series))
        is_dask_array_X = isinstance(X, da.Array)

        if is_dask_frame_X:
            npartitions = X.npartitions
            X = X.compute()
        elif is_dask_array_X:
            chunks = X.chunks
            X = X.compute()

        args = [a.compute() if isinstance(a, (dd.DataFrame, dd.Series, da.Array)) else a for a in args]
        r = fn(X, *args, **kwargs)

        if isinstance(r, (pd.DataFrame, pd.Series, np.ndarray)):
            if is_dask_frame_X:
                r = dd.from_pandas(r, npartitions=npartitions)
            elif is_dask_array_X:
                r = da.from_array(r, chunks=chunks)

        return r

    def __getattribute__(self, name: str):
        if name in {'adapted_attributes_', 'adapted_transformer_'}:
            return super().__getattribute__(name)

        adapted_attributes = self.adapted_attributes_
        if adapted_attributes is not None and name in adapted_attributes:
            adapted_transformer = self.adapted_transformer_
            return getattr(adapted_transformer, name)

        return super().__getattribute__(name)

    def __dir__(self):
        adapted_attributes = self.adapted_attributes_
        if adapted_attributes is not None:
            return set(list(adapted_attributes) + list(super().__dir__()))

        return super().__dir__()


class LgbmLeavesEncoder(AdaptedTransformer, TransformerMixin):

    def __init__(self, cat_vars, cont_vars, task, **params):
        attributes = {'lgbm', 'cat_vars', 'cont_vars', 'new_columns', 'task', 'lgbm_params'}
        super(LgbmLeavesEncoder, self).__init__(skex.LgbmLeavesEncoder, attributes,
                                                cat_vars, cont_vars, task, **params)


class CategorizeEncoder(skex.CategorizeEncoder):

    def fit(self, X, y=None):
        super(CategorizeEncoder, self).fit(X, y)

        if len(self.new_columns) > 0:
            nuniques = [nunique for _, _, nunique in self.new_columns]
            nuniques = dask.compute(*nuniques)

            new_columns = [(col, dtype, nunique)
                           for (col, dtype, _), nunique in zip(self.new_columns, nuniques)]
            self.new_columns = new_columns

        return self


class MultiKBinsDiscretizer(BaseEstimator, TransformerMixin):
    def __init__(self, columns=None, bins=None, strategy='quantile', dtype='float64'):
        assert strategy in {'uniform', 'quantile'}
        assert dtype is None or dtype.startswith('int') or dtype.startswith('float')

        super(MultiKBinsDiscretizer, self).__init__()

        self.columns = columns
        self.bins = bins
        self.strategy = strategy
        self.dtype = dtype

        # fitted
        self.new_columns = []
        self.bin_edges_ = {}

    def fit(self, X, y=None):
        assert isinstance(X, (dd.DataFrame, pd.DataFrame))
        is_dask_X = isinstance(X, dd.DataFrame)

        new_columns = []
        if self.columns is None:
            self.columns = X.select_dtypes(['float', 'float64', 'int', 'int64']).columns.tolist()

            # define new_columns
        for col in self.columns:
            new_name = col + const.COLUMNNAME_POSTFIX_DISCRETE
            c_bins = self.bins
            if c_bins is None or c_bins <= 0:
                n_unique = X.loc[:, col].nunique()
                if is_dask_X:
                    n_unique = n_unique.compute()
                c_bins = int(round(n_unique ** 0.25)) + 1
            new_columns.append((col, new_name, c_bins))

        # compute bin_borders
        if self.strategy == 'quantile':
            bin_edges = {}
            for col, _, c_bins in new_columns:
                step = 1.0 / c_bins
                qs = [i * step for i in range(1, int(c_bins))]
                q = X[col].quantile(qs)
                if is_dask_X:
                    q = q.compute()
                bin_edges[col] = q.to_list()
        else:  # strategy == 'uniform'
            X = X[self.columns]
            mns, mxs = X.min(), X.max()
            if is_dask_X:
                mns, mxs = dask.compute(mns, mxs)
            bin_edges = {col: [mns[col], mxs[col]] for col in self.columns}

        self.new_columns = new_columns
        self.bin_edges_ = bin_edges

        return self

    def transform(self, X):
        assert isinstance(X, (dd.DataFrame, pd.DataFrame))

        if self.strategy == 'quantile':
            transformer = MultiKBinsDiscretizer._transform_df_quantile
        else:  # strategy == 'uniform'
            transformer = MultiKBinsDiscretizer._transform_df_uniform

        if isinstance(X, dd.DataFrame):
            meta = X.dtypes.to_dict()
            dtype = self.dtype if self.dtype is not None else 'int'
            for _, new_name, _ in self.new_columns:
                meta[new_name] = dtype
            fn = partial(transformer, self.new_columns, self.bin_edges_, self.dtype)
            X = X.map_partitions(fn, meta=meta)
        else:
            X = transformer(self.new_columns, self.bin_edges_, self.dtype, X)

        return X

    @staticmethod
    def _transform_df_quantile(new_columns, bin_edges, dtype, X):
        for col, new_name, c_bins in new_columns:
            edges = bin_edges[col]
            X[new_name] = np.searchsorted(edges, X[col], side='left')

            if dtype is not None and not str(dtype).startswith('int'):
                X[new_name] = X[new_name].astype(dtype)

        return X

    @staticmethod
    def _transform_df_uniform(new_columns, bin_edges, dtype, X):
        for col, new_name, c_bins in new_columns:
            mn, mx = bin_edges[col]
            if dtype is not None and str(dtype).startswith('float'):
                if mx > mn and c_bins > 1:
                    step = (mx - mn) / c_bins
                    fn = lambda v: 0. if v <= mn else c_bins - 1. if v >= mx else float(int((v - mn) / step))
                else:
                    fn = lambda v: float(v > mn)
            else:
                if mx > mn and c_bins > 1:
                    step = (mx - mn) / c_bins
                    fn = lambda v: 0 if v <= mn else c_bins - 1 if v >= mx else int((v - mn) / step)
                else:
                    fn = lambda v: int(v > mn)

            X[new_name] = X[col].map(fn)

        return X


class MultiVarLenFeatureEncoder(BaseEstimator, TransformerMixin):
    """
    Example:
        features = [ ('feature_1','|'), ... ]
        encoder = MultiVarLenFeatureEncoder(features)
        ...
    """

    def __init__(self, features):
        assert isinstance(features, (tuple, list))
        assert all([isinstance(fs, (tuple, list)) and len(fs) > 1 for fs in features])
        assert all([f[0] for f in features]), 'Feature name is required.'
        assert all([f[1] for f in features]), 'Separator is required.'

        super(MultiVarLenFeatureEncoder, self).__init__()

        self.seps_ = {feature[0]: feature[1] for feature in features}

        # fitted
        self.keys_ = None
        self.max_length_ = {}  # key: feature, value: max length

    def fit(self, X, y=None):
        assert isinstance(X, (dd.DataFrame, pd.DataFrame))

        is_dask_X = isinstance(X, dd.DataFrame)
        feature_keys = {}
        max_length = {}

        sep_to_features = defaultdict(list)
        for f, sep in self.seps_.items():
            sep_to_features[sep].append(f)

        for sep, features in sep_to_features.items():
            if is_dask_X:
                meta = {f: 'str' for f in features}
                fn_chunk = partial(MultiVarLenFeatureEncoder._fit_part, sep)
                fn_agg = partial(MultiVarLenFeatureEncoder._agg_part, sep)
                t = X[features].reduction(fn_chunk, aggregate=fn_agg, meta=meta,
                                          token=f'{self.__class__.__name__}.fit'
                                          ).compute()
            else:  # pd.DataFrame
                t = self._fit_part(sep, X[features])

            assert isinstance(t, pd.Series)
            for f, aggregated in t.to_dict().items():
                a = aggregated.split(sep)
                max_length[f] = int(a[0])
                feature_keys[f] = a[1:]

        self.max_length_ = max_length
        self.keys_ = feature_keys

        return self

    def transform(self, X):
        assert isinstance(X, (dd.DataFrame, pd.DataFrame))
        assert len(set(self.seps_.keys()) - set(X.columns.to_list())) == 0, \
            f'Not found {set(self.seps_.keys()) - set(X.columns.to_list())} in X'

        if isinstance(X, dd.DataFrame):
            meta = X.dtypes.to_dict()
            for f in self.seps_.keys():
                meta[f] = 'object'
            fn = partial(MultiVarLenFeatureEncoder._transform_df, self.seps_, self.keys_, self.max_length_)
            X = X.map_partitions(fn, meta=meta)
        else:
            X = self._transform_df(self.seps_, self.keys_, self.max_length_, X)

        return X

    @staticmethod
    def _fit_part(sep, X):
        assert isinstance(X, pd.DataFrame)
        fn = partial(MultiVarLenFeatureEncoder._aggregate_key, sep, False)
        return X.agg(fn, axis=0)

    @staticmethod
    def _agg_part(sep, X):
        assert isinstance(X, pd.DataFrame)
        fn = partial(MultiVarLenFeatureEncoder._aggregate_key, sep, True)
        return X.agg(fn, axis=0)

    @staticmethod
    def _aggregate_key(sep, y_include_max_len, y):
        key_set = set()
        max_len = 0

        for v in y:
            a = list(filter(len, v.split(sep)))
            if y_include_max_len:
                len_a = int(a[0])
                a = a[1:]
            else:
                len_a = len(a)
            if len_a > max_len:
                max_len = len_a
            key_set.update(a)
        key_set = list(key_set)
        key_set.sort()

        return sep.join([str(max_len)] + key_set)

    @staticmethod
    def _transform_df(feature_seps, feature_keys, feature_max_length, X):
        for f, sep in feature_seps.items():
            keys = feature_keys.get(f, [])
            max_len = feature_max_length.get(f, 0)
            assert isinstance(keys, (list, tuple)) and isinstance(max_len, int)
            if len(keys) <= 1 or max_len < 1:
                continue
            X[f] = MultiVarLenFeatureEncoder._encode(sep, keys, max_len, X[f])

        return X

    @staticmethod
    def _encode(sep, keys, max_len, y):
        unseen = len(keys) + 1
        key_values = dict(zip(keys, list(range(1, unseen + 1))))
        result = np.empty(len(y), dtype=np.object)

        for yi, v in enumerate(y):
            result_yi = []
            arr = filter(len, v.split(sep))
            for vi, k in enumerate(arr):
                result_yi.append(key_values[k] if k in key_values else unseen)
                if vi >= max_len - 1:
                    break
            while len(result_yi) < max_len:
                result_yi.append(0)
            result[yi] = result_yi

        return result


class LocalizedTfidfVectorizer(BaseEstimator, TransformerMixin):
    def __init__(self, max_features=None, max_df=1.0, min_df=1):
        super(LocalizedTfidfVectorizer, self).__init__()

        self.max_features = max_features
        self.max_df = max_df
        self.min_df = min_df

        # fitted
        self.vocabulary_ = None

    def fit(self, X, y=None, **kwargs):
        chunk_kwargs = dict(max_features=self.max_features, max_df=self.max_df, min_df=self.min_df)

        if isinstance(X, dd.Series):
            voc_df = X.reduction(self.fit_part, aggregate=self.agg_part, chunk_kwargs=chunk_kwargs, meta=(None, 'f8'))
            voc_df = voc_df.compute()
        elif isinstance(X, (pd.Series, np.ndarray)):
            voc_df = self.fit_part(X, **chunk_kwargs)
        else:
            raise ValueError(f'Unsupported data type: {type(X).__name__}')

        vocabulary = {v: i for v, i in zip(voc_df.index, range(len(voc_df)))}
        self.vocabulary_ = vocabulary

        return self

    def transform(self, X, y=None):
        if isinstance(X, dd.Series):
            meta = {f'x{i}': 'f8' for _, i in self.vocabulary_.items()}
            r = X.map_partitions(self.transform_part, self.vocabulary_, meta=meta)
        elif isinstance(X, (pd.Series, np.ndarray)):
            r = self.transform_part(X, self.vocabulary_)
        else:
            raise ValueError(f'Unsupported data type: {type(X).__name__}')

        return r

    @staticmethod
    def fit_part(part, max_features=None, max_df=None, min_df=None):
        t = skex.LocalizedTfidfVectorizer(use_idf=False, max_features=max_features, max_df=max_df, min_df=min_df)
        t.fit(part)
        return pd.Series(t.vocabulary_)

    @staticmethod
    def agg_part(part):
        return pd.Series(index=part.columns)

    @staticmethod
    def transform_part(part, voc):
        t = skex.LocalizedTfidfVectorizer(use_idf=False, vocabulary=voc)
        t.fit(np.array(['']))
        result = t.transform(part).toarray()

        if isinstance(part, pd.Series):
            columns = [f'x{i}' for _, i in voc.items()]
            result = pd.DataFrame(result, index=part.index, columns=columns)

        return result



import time

from hypernets.dispatchers.predict.grpc.proto import predict_pb2_grpc
from hypernets.dispatchers.predict.grpc.proto.predict_pb2 import PredictResponse
from hypernets.dispatchers.process import LocalProcess
from hypernets.utils import logging

logger = logging.get_logger(__name__)


class PredictService(predict_pb2_grpc.PredictServiceServicer):
    def __init__(self, cmd):
        super(PredictService, self).__init__()
        assert cmd

        self.cmd = cmd

    def predict(self, request, context):
        data_file = request.data_file
        result_file = request.result_file

        start_at = time.time()

        if logger.is_info_enabled():
            print(f'predict {data_file} --> {result_file}', end='')

        cmd = f'{self.cmd} {data_file} {result_file}'
        p = LocalProcess(cmd, None, None, None)
        p.start()
        p.join()
        code = p.exitcode

        res = PredictResponse(data_file=data_file, result_file=result_file, code=code)

        done_at = time.time()
        if logger.is_info_enabled():
            print(' done, elapsed %.3f seconds.' % (done_at - start_at))
        return res


def serve(addr, cmd):
    import grpc
    from concurrent import futures

    if logger.is_info_enabled():
        logger.info(f'start predict service at {addr}')
    service = PredictService(cmd)
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    predict_pb2_grpc.add_PredictServiceServicer_to_server(service, server)

    server.add_insecure_port(addr)
    server.start()

    return server, service



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
from hypernets.discriminators import PercentileDiscriminator, get_previous_trials_scores, get_percentile_score
from hypernets.core import TrialHistory, Trial

history = TrialHistory(optimize_direction='min')
group_id = 'lightgbm_cv_1'
group_id2 = 'lightgbm_cv_2'
t1 = Trial(None, 1, 0.9, 0, succeeded=True)
t1.iteration_scores[group_id] = [0.9, 0.8, 0.7, 0.6, 0.5, 0.4, 0.3, 0.3, 0.3]
t2 = Trial(None, 1, 0.8, 0, succeeded=True)
t2.iteration_scores[group_id] = [0.9, 0.85, 0.75, 0.65, 0.55, 0.45, 0.35, 0.35, 0.35, 0.25]
t3 = Trial(None, 1, 0.8, 0, succeeded=True)
t3.iteration_scores[group_id] = [0.9, 0.85, 0.75, 0.65, 0.54, 0.44, 0.34, 0.34, 0.34, 0.24]
t4 = Trial(None, 1, 0.8, 0, succeeded=True)
t4.iteration_scores[group_id] = [0.9, 0.85, 0.75, 0.65, 0.53, 0.43, 0.33, 0.33, 0.33, 0.23]
t5 = Trial(None, 1, 0.8, 0, succeeded=True)
t5.iteration_scores[group_id] = [0.9, 0.85, 0.75, 0.65, 0.52, 0.42, 0.32, 0.32, 0.32, 0.22]
t6 = Trial(None, 1, 0.8, 0, succeeded=True)
t6.iteration_scores[group_id] = [0.9, 0.85, 0.75, 0.65, 0.51, 0.41, 0.31, 0.31, 0.31, 0.21]
t6.iteration_scores[group_id2] = [0.9, 0.85, 0.75, 0.65, 0.51, 0.41, 0.31, 0.31, 0.31, 0.21]
history.append(t1)
history.append(t2)
history.append(t3)
history.append(t4)
history.append(t5)
history.append(t6)


# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
import numpy as np

_hypernets_random_state = None


def set_random_state(seed):
    global _hypernets_random_state
    if seed is None:
        _hypernets_random_state = None
    else:
        _hypernets_random_state = np.random.RandomState(seed=seed)


def get_random_state():
    global _hypernets_random_state
    if _hypernets_random_state is None:
        return np.random.RandomState()
    else:
        return _hypernets_random_state


def randint():
    return get_random_state().randint(0, 65535)



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
import numpy as np

from ..core import TrialHistory
from ..utils import to_repr


class UnPromisingTrial(Exception):
    pass


class BaseDiscriminator(object):
    """
    Discriminator is used to determine whether to continue training
    """

    def __init__(self, min_trials=5, min_steps=5, stride=1, history: TrialHistory = None, optimize_direction='min'):
        super(BaseDiscriminator, self).__init__()

        self.history = history
        self.min_trials = min_trials
        self.min_steps = min_steps
        self.stride = stride
        self.optimize_direction = optimize_direction.lower()
        if self.optimize_direction == 'min':
            self._sign = -1
        else:
            self._sign = 1

    def bind_history(self, history):
        self.history = history

    def is_promising(self, iteration_trajectory, group_id):
        if self.history is None:
            raise ValueError('`history` is not bound')
        n_step = len(iteration_trajectory)
        if n_step < self.min_steps:
            return True

        trial_scores = get_previous_trials_scores(self.history, n_step - 1, n_step - 1, group_id)
        if len(trial_scores) < self.min_trials:
            return True
        if self.stride > 1:
            if ((n_step - self.min_steps) % self.stride) > 0:
                return True

        return self._is_promising(iteration_trajectory, group_id)

    def _is_promising(self, iteration_trajectory, group_id):
        """
        discriminate whether continuing training a trial is promising

        Args:
            iteration_trajectory:
                list, The scores of each step in the iteration process are arranged from front to back
            group_id:
                Str, It is used to group different types of trials in a search task
        Returns:
            A boolean value representing whether the trial should be stop.
        """
        raise NotImplementedError()

    def __repr__(self):
        return to_repr(self)


def get_percentile_score(history, n_step, group_id, percentile, sign=1):
    trial_scores = get_previous_trials_scores(history, n_step, n_step, group_id)
    percentile_score = np.percentile(trial_scores * sign, percentile) * sign
    return percentile_score


def get_previous_trials_scores(history, from_step, to_step, group_id):
    trial_scores = []
    for trial in history.trials:
        if not trial.succeeded:
            continue
        scores = trial.iteration_scores.get(group_id)
        if scores and len(scores) >= (to_step + 1):
            trial_scores.append(scores[from_step:(to_step + 1)])
    return np.array(trial_scores)



# -*- coding:utf-8 -*-
"""

"""
import time
import traceback
from collections import UserDict

from ..core.meta_learner import MetaLearner
from ..core.trial import *
from ..discriminators import UnPromisingTrial
from ..dispatchers import get_dispatcher
from ..utils import logging, infer_task_type as _infer_task_type, hash_data, const, to_repr

logger = logging.get_logger(__name__)


class HyperModel:
    def __init__(self, searcher, dispatcher=None, callbacks=None, reward_metric=None, task=None, discriminator=None):
        """

        :param searcher:
        :param dispatcher:
        :param callbacks:
        :param reward_metric:
        :param task:
        """
        self.searcher = searcher
        self.dispatcher = dispatcher
        self.callbacks = callbacks if callbacks is not None else []
        self.reward_metric = reward_metric
        self.history = TrialHistory(searcher.optimize_direction)
        self.task = task
        self.discriminator = discriminator
        if self.discriminator:
            self.discriminator.bind_history(self.history)

    def _get_estimator(self, space_sample):
        raise NotImplementedError

    def load_estimator(self, model_file):
        raise NotImplementedError

    def _run_trial(self, space_sample, trial_no, X, y, X_eval, y_eval, cv=False, num_folds=3, model_file=None,
                   **fit_kwargs):

        start_time = time.time()
        estimator = self._get_estimator(space_sample)
        if self.discriminator:
            estimator.set_discriminator(self.discriminator)

        for callback in self.callbacks:
            callback.on_build_estimator(self, space_sample, estimator, trial_no)
        succeeded = False
        scores = None
        oof = None
        oof_scores = None
        try:
            if cv:
                scores, oof, oof_scores = estimator.fit_cross_validation(X, y, stratified=True, num_folds=num_folds,
                                                                         shuffle=False, random_state=9527,
                                                                         metrics=[self.reward_metric],
                                                                         **fit_kwargs)
            else:
                estimator.fit(X, y, **fit_kwargs)
            succeeded = True
        except UnPromisingTrial as e:
            logger.info(f'{e}')
        except Exception as e:
            logger.error(f'run_trail failed! trail_no={trial_no}')
            track = traceback.format_exc()
            logger.error(track)

        if succeeded:
            if scores is None:
                scores = estimator.evaluate(X_eval, y_eval, metrics=[self.reward_metric], **fit_kwargs)
            reward = self._get_reward(scores, self.reward_metric)

            if model_file is None or len(model_file) == 0:
                model_file = '%05d_%s.pkl' % (trial_no, space_sample.space_id)
            estimator.save(model_file)

            elapsed = time.time() - start_time

            trial = Trial(space_sample, trial_no, reward, elapsed, model_file, succeeded)
            trial.iteration_scores = estimator.get_iteration_scores()
            if oof is not None:
                trial.memo['oof'] = oof
            if oof_scores is not None:
                trial.memo['oof_scores'] = oof_scores

            # improved = self.history.append(trial)

            self.searcher.update_result(space_sample, reward)
        else:
            elapsed = time.time() - start_time
            trial = Trial(space_sample, trial_no, 0, elapsed, succeeded=succeeded)

            if self.history is not None:
                t = self.history.get_worst()
                if t is not None:
                    self.searcher.update_result(space_sample, t.reward)

        return trial

    def _get_reward(self, value, key=None):
        def cast_float(value):
            try:
                fv = float(value)
                return fv
            except TypeError:
                return None

        if key is None:
            key = 'reward'

        fv = cast_float(value)
        if fv is not None:
            reward = fv
        elif (isinstance(value, dict) or isinstance(value, UserDict)) and key in value and cast_float(
                value[key]) is not None:
            reward = cast_float(value[key])
        else:
            raise ValueError(
                f'[value] should be a numeric or a dict which has a key named "{key}" whose value is a numeric.')
        return reward

    def get_best_trial(self):
        return self.history.get_best()

    @property
    def best_reward(self):
        best = self.get_best_trial()
        if best is not None:
            return best.reward
        else:
            return None

    @property
    def best_trial_no(self):
        best = self.get_best_trial()
        if best is not None:
            return best.trial_no
        else:
            return None

    def get_top_trials(self, top_n):
        return self.history.get_top(top_n)

    def _before_search(self):
        pass

    def _after_search(self, last_trial_no):
        pass

    def search(self, X, y, X_eval, y_eval, cv=False, num_folds=3, max_trials=10, dataset_id=None, trial_store=None,
               **fit_kwargs):
        """
        :param X: Pandas or Dask DataFrame, feature data for training
        :param y: Pandas or Dask Series, target values for training
        :param X_eval: (Pandas or Dask DataFrame) or None, feature data for evaluation
        :param y_eval: (Pandas or Dask Series) or None, target values for evaluation
        :param cv: Optional, int(default=False), If set to `true`, use cross-validation instead of evaluation set reward to guide the search process
        :param num_folds: Optional, int(default=3), Number of cross-validated folds, only valid when cv is true
        :param max_trials: Optional, int(default=10), The upper limit of the number of search trials, the search process stops when the number is exceeded
        :param dataset_id:
        :param trial_store:
        :param fit_kwargs: Optional, dict, parameters for fit method of model
        :return:
        """
        if self.task is None or self.task == const.TASK_AUTO:
            self.task, _ = self.infer_task_type(y)
        if self.task not in [const.TASK_BINARY, const.TASK_MULTICLASS, const.TASK_REGRESSION, const.TASK_MULTILABEL]:
            logger.warning(f'Unexpected task "{self.task}"')

        if dataset_id is None:
            dataset_id = self.generate_dataset_id(X, y)
        if self.searcher.use_meta_learner:
            self.searcher.set_meta_learner(MetaLearner(self.history, dataset_id, trial_store))

        self._before_search()

        dispatcher = self.dispatcher if self.dispatcher else get_dispatcher(self)

        for callback in self.callbacks:
            callback.on_search_start(self, X, y, X_eval, y_eval,
                                     cv, num_folds, max_trials, dataset_id, trial_store,
                                     **fit_kwargs)
        try:
            trial_no = dispatcher.dispatch(self, X, y, X_eval, y_eval,
                                           cv, num_folds, max_trials, dataset_id, trial_store,
                                           **fit_kwargs)

            for callback in self.callbacks:
                callback.on_search_end(self)
        except Exception as e:
            for callback in self.callbacks:
                callback.on_search_error(self)
            raise e

        self._after_search(trial_no)

    def generate_dataset_id(self, X, y):
        if hasattr(X, 'shape') and len(getattr(X, 'shape')) == 2:
            sign = hash_data([X, y])
        else:
            import hashlib
            repr = ''
            if X is not None:
                if isinstance(X, list):
                    repr += f'X len({len(X)})|'
                if hasattr(X, 'shape'):
                    repr += f'X shape{X.shape}|'
                if hasattr(X, 'dtypes'):
                    repr += f'x.dtypes({list(X.dtypes)})|'

            if y is not None:
                if isinstance(y, list):
                    repr += f'y len({len(y)})|'
                if hasattr(y, 'shape'):
                    repr += f'y shape{y.shape}|'

                if hasattr(y, 'dtype'):
                    repr += f'y.dtype({y.dtype})|'

            sign = hashlib.md5(repr.encode('utf-8')).hexdigest()
        return sign

    def final_train(self, space_sample, X, y, **kwargs):
        estimator = self._get_estimator(space_sample)
        estimator.set_discriminator(None)
        estimator.fit(X, y, **kwargs)
        return estimator

    def export_configuration(self, trials):
        configurations = []
        for trial in trials:
            configurations.append(self.export_trial_configuration(trial))
        return configurations

    def export_trial_configuration(self, trial):
        raise NotImplementedError

    def infer_task_type(self, y):
        return _infer_task_type(y)

    def plot_hyperparams(self, destination='notebook', output='hyperparams.html'):
        return self.history.plot_hyperparams(destination, output)

    def __repr__(self):
        return to_repr(self)



# -*- coding:utf-8 -*-
import argparse

from hypernets.dispatchers.process.grpc.process_broker_service import serve


def main():
    parser = argparse.ArgumentParser('run HyperNets process broker.')
    parser.add_argument('--host', '-host',
                        default='0.0.0.0',
                        help='broker hostname or ip address'
                             + ', default "0.0.0.0"')
    parser.add_argument('--port', '-port',
                        type=int, default=8010,
                        help='broker tcp port, default 8010')
    parser.add_argument('--workers', '-workers',
                        type=int, default=10,
                        help='max worker count, default 10')

    args = parser.parse_args()

    server, _ = serve(f'{args.host}:{args.port}', args.workers)
    server.wait_for_termination()


if __name__ == '__main__':
    try:
        main()
        print('done')
    except KeyboardInterrupt as e:
        print(e)



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from collections import defaultdict

import numpy as np
import pandas as pd
from scipy.cluster import hierarchy
from scipy.stats import spearmanr
from sklearn.impute import SimpleImputer

from hypernets.core import randint
from hypernets.tabular import sklearn_ex as skex, dask_ex as dex, column_selector as cs
from hypernets.tabular.cfg import TabularCfg as cfg
from hypernets.utils import logging

logger = logging.get_logger(__name__)


def select_by_multicollinearity(X, method=None):
    """
    Adapted from https://scikit-learn.org/stable/auto_examples/inspection/plot_permutation_importance_multicollinear.html
    handling multicollinearity is by performing hierarchical clustering on the features’ Spearman
    rank-order correlations, picking a threshold, and keeping a single feature from each cluster.
    """
    X_shape = X.shape
    if dex.is_dask_dataframe(X):
        X_shape = dex.compute(X_shape)[0]
    sample_limit = cfg.multi_collinearity_sample_limit
    if X_shape[0] > sample_limit:
        logger.info(f'{X_shape[0]} rows data found, sample to {sample_limit}')
        frac = sample_limit / X_shape[0]
        X, _, = dex.train_test_split(X, train_size=frac, random_state=randint())

    n_values = [X[c].value_counts() for c in X.columns]
    if dex.is_dask_dataframe(X):
        n_values = dex.compute(*n_values)
    one_values = [n.name for n in n_values if len(n) <= 1]
    if len(one_values) > 0:
        X = X[[c for c in X.columns if c not in one_values]]

    logger.info('computing correlation')
    if (method is None or method == 'spearman') and isinstance(X, pd.DataFrame):
        Xt = SimpleImputer(missing_values=np.nan, strategy='most_frequent').fit_transform(X)
        corr = spearmanr(Xt).correlation
    elif isinstance(X, pd.DataFrame):
        Xt = X.copy()
        cols = cs.column_number_exclude_timedelta(X)
        if cols:
            Xt[cols] = SimpleImputer(missing_values=np.nan, strategy='most_frequent').fit_transform(Xt[cols])
        Xt = skex.SafeOrdinalEncoder().fit_transform(Xt)
        corr = Xt.corr(method=method).values
    else:  # dask
        Xt = dex.SafeOrdinalEncoder().fit_transform(X)
        corr = Xt.corr(method='pearson' if method is None else method).compute().values

    logger.info('computing cluster')
    corr_linkage = hierarchy.ward(corr)
    cluster_ids = hierarchy.fcluster(corr_linkage, 1, criterion='distance')
    cluster_id_to_feature_ids = defaultdict(list)
    for idx, cluster_id in enumerate(cluster_ids):
        cluster_id_to_feature_ids[cluster_id].append(idx)
    selected = [X.columns[v[0]] for v in cluster_id_to_feature_ids.values()]
    unselected = list(set(X.columns.to_list()) - set(selected)) + one_values
    feature_clusters = [[X.columns[i] for i in v] for v in cluster_id_to_feature_ids.values()]
    return feature_clusters, selected, unselected



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from ._base import get_previous_trials_scores, get_percentile_score, UnPromisingTrial, BaseDiscriminator
from .percentile import PercentileDiscriminator, ProgressivePercentileDiscriminator

_discriminators = {
    'percentile': PercentileDiscriminator,
    'percentile_discriminator': PercentileDiscriminator,

    'progressive': ProgressivePercentileDiscriminator,
    'progressive_percentile': ProgressivePercentileDiscriminator,
    'progressive_percentile_discriminator': ProgressivePercentileDiscriminator,
}


def _get_discriminator_cls(identifier):
    if isinstance(identifier, str):
        cls = _discriminators.get(identifier.lower(), None)
        if cls is not None:
            return cls
    elif isinstance(identifier, type) and issubclass(identifier, BaseDiscriminator):
        return identifier

    raise ValueError(f'Illegal discriminator:{identifier}')


def make_discriminator(cls, optimize_direction='min', **kwargs):
    cls = _get_discriminator_cls(cls)

    if cls == PercentileDiscriminator:
        default_kwargs = dict(percentile=0)
    elif cls == ProgressivePercentileDiscriminator:
        default_kwargs = dict(percentile_list=[0])
    else:
        default_kwargs = {}

    kwargs = {**default_kwargs, **kwargs}
    discriminator = cls(optimize_direction=optimize_direction, **kwargs)
    return discriminator



import copy
import pickle

from sklearn.pipeline import Pipeline

from . import fs


def save_estimator(estimator, model_path):
    if isinstance(estimator, Pipeline) and hasattr(estimator.steps[-1][1], 'save') \
            and hasattr(estimator.steps[-1][1], 'load'):
        if fs.exists(model_path):
            fs.rm(model_path, recursive=True)
        fs.mkdirs(model_path, exist_ok=True)
        if not model_path.endswith(fs.sep):
            model_path = model_path + fs.sep

        stub = copy.copy(estimator)
        stub.steps[-1][1].save(f'{model_path}pipeline.model')
        with fs.open(f'{model_path}pipeline.pkl', 'wb') as f:
            pickle.dump(stub, f, protocol=pickle.HIGHEST_PROTOCOL)
    else:
        with fs.open(model_path, 'wb') as f:
            pickle.dump(estimator, f, protocol=pickle.HIGHEST_PROTOCOL)


def load_estimator(model_path):
    model_path_ = model_path
    if not model_path_.endswith(fs.sep):
        model_path_ = model_path_ + fs.sep

    if fs.exists(f'{model_path_}pipeline.pkl'):
        with fs.open(f'{model_path_}pipeline.pkl', 'rb') as f:
            stub = pickle.load(f)
            assert isinstance(stub, Pipeline)

        estimator = stub.steps[-1][1]
        if fs.exists(f'{model_path_}pipeline.model') and hasattr(estimator, 'load'):
            est = estimator.load(f'{model_path_}pipeline.model')
            steps = stub.steps[:-1] + [(stub.steps[-1][0], est)]
            stub = Pipeline(steps)
    else:
        with fs.open(model_path, 'rb') as f:
            stub = pickle.load(f)

    return stub



# -*- coding:utf-8 -*-
import argparse

from hypernets.dispatchers.predict.grpc.predict_service import serve


def main():
    parser = argparse.ArgumentParser('start predict server.')
    parser.add_argument('--port', '-port',
                        type=int, default=8030,
                        help='tcp port of the predict server')
    args, argv = parser.parse_known_args()

    server, _ = serve(f'0.0.0.0:{args.port}', ' '.join(argv))
    server.wait_for_termination()


if __name__ == '__main__':
    try:
        main()
        print('done')
    except KeyboardInterrupt as e:
        print(e)



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

import hashlib
import inspect
import math
import pickle
import uuid
from collections import OrderedDict
from functools import partial
from io import BytesIO

import dask.array as da
import dask.dataframe as dd
import numpy as np
import pandas as pd

from . import logging
from .const import TASK_BINARY, TASK_REGRESSION, TASK_MULTICLASS, TASK_MULTILABEL

logger = logging.get_logger(__name__)


def generate_id():
    return str(uuid.uuid1())


def get_params(obj, include_default=False):
    def _get_init_params(cls):
        init = cls.__init__
        if init is object.__init__:
            return []

        init_signature = inspect.signature(init)
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self' and p.kind != p.VAR_KEYWORD]
        return parameters

    out = OrderedDict()
    for p in _get_init_params(type(obj)):
        name = p.name
        value = getattr(obj, name, None)
        if include_default or value is not p.default:
            out[name] = value

    return out


def to_repr(obj, excludes=None):
    try:
        if excludes is None:
            excludes = []
        out = ['%s=%r' % (k, v) for k, v in get_params(obj).items() if k not in excludes]
        repr_ = ', '.join(out)
        return f'{type(obj).__name__}({repr_})'
    except Exception as e:
        if logger.is_info_enabled():
            logger.info(e)
        return f'{e} <to_repr>: {obj}'


def combinations(n, m_max, m_min=1):
    if m_max > n or m_max <= 0:
        m_max = n
    if m_min < 1:
        m_min = 1
    if m_min == 1 and m_max == n:
        return 2 ** n - 1
    else:
        sum = 0
        for i in range(m_min, m_max + 1):
            c = math.factorial(n) / (math.factorial(i) * math.factorial(n - i))
            sum += c
        return sum


class Counter(object):
    def __init__(self):
        from threading import Lock

        super(Counter, self).__init__()
        self._value = 0
        self._lock = Lock()

    @property
    def value(self):
        return self._value

    def __call__(self, *args, **kwargs):
        with self._lock:
            self._value += 1
            return self._value

    def inc(self, step=1):
        with self._lock:
            self._value += step
            return self._value

    def reset(self):
        with self._lock:
            self._value = 0
            return self._value


def isnotebook():
    '''code from https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
    :return:
    '''
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except:
        return False


def infer_task_type(y, excludes=None):
    assert excludes is None or isinstance(excludes, (list, tuple, set))

    if len(y.shape) > 1 and y.shape[-1] > 1:
        labels = list(range(y.shape[-1]))
        task = TASK_MULTILABEL  # 'multilable'
        return task, labels

    if hasattr(y, 'unique'):
        uniques = set(y.unique())
    elif isinstance(y, da.Array):
        uniques = set(da.unique(y).compute())
    else:
        uniques = set(y)

    if uniques.__contains__(np.nan):
        uniques.remove(np.nan)
    if excludes is not None and len(excludes) > 0:
        uniques -= set(excludes)
    n_unique = len(uniques)
    labels = []

    if n_unique == 2:
        logger.info(f'2 class detected, {uniques}, so inferred as a [binary classification] task')
        task = TASK_BINARY  # TASK_BINARY
        labels = sorted(uniques)
    else:
        if str(y.dtype).find('float') >= 0:
            logger.info(f'Target column type is {y.dtype}, so inferred as a [regression] task.')
            task = TASK_REGRESSION
        else:
            if n_unique > 1000:
                if str(y.dtype).find('int') >= 0:
                    logger.info('The number of classes exceeds 1000 and column type is {y.dtype}, '
                                'so inferred as a [regression] task ')
                    task = TASK_REGRESSION
                else:
                    raise ValueError('The number of classes exceeds 1000, please confirm whether '
                                     'your predict target is correct ')
            else:
                logger.info(f'{n_unique} class detected, inferred as a [multiclass classification] task')
                task = TASK_MULTICLASS
                labels = sorted(uniques)
    return task, labels


def hash_dataframe(df, method='md5', index=False):
    assert isinstance(df, (pd.DataFrame, dd.DataFrame))

    m = getattr(hashlib, method)()

    for col in df.columns:
        m.update(str(col).encode())

    if isinstance(df, dd.DataFrame):
        x = df.map_partitions(lambda part: pd.util.hash_pandas_object(part, index=index),
                              meta=(None, 'u8')).compute()
    else:
        x = pd.util.hash_pandas_object(df, index=index)

    np.vectorize(m.update, otypes=[None], signature='()->()')(x.values)

    return m.hexdigest()


def _hash_array(arr):
    if arr.shape[0] == 0:
        v = np.array([], dtype='u8').reshape((-1, 1))
    else:
        v = pd.util.hash_pandas_object(pd.DataFrame(arr), index=False).values.reshape((-1, 1))

    return v


def hash_array(arr, method='md5'):
    m = getattr(hashlib, method)()

    if isinstance(arr, da.Array):
        if len(arr.shape) == 1:
            arr = arr.compute_chunk_sizes().reshape(-1, 1)
        x = arr.map_blocks(_hash_array, dtype='u8').compute()
    elif isinstance(arr, np.ndarray):
        x = _hash_array(arr)
    else:
        x = _hash_array(np.array(arr))

    np.vectorize(m.update, otypes=[None], signature='()->()')(x)

    return m.hexdigest()


def hash_data(data, method='md5'):
    if isinstance(data, (pd.DataFrame, dd.DataFrame)):
        return hash_dataframe(data, method=method)
    elif isinstance(data, (pd.Series, dd.Series)):
        return hash_dataframe(data.to_frame(), method=method)
    elif isinstance(data, (np.ndarray, da.Array)):
        return hash_array(data, method=method)

    if isinstance(data, (bytes, bytearray)):
        pass
    elif isinstance(data, str):
        data = data.encode('utf-8')
    else:
        if isinstance(data, (list, tuple)):
            data = [hash_data(x) if x is not None else x for x in data]
        elif isinstance(data, dict):
            data = {hash_data(k): hash_data(v) if v is not None else v for k, v in data.items()}
        buf = BytesIO()
        pickle.dump(data, buf)
        data = buf.getvalue()
        buf.close()

    m = getattr(hashlib, method)()
    m.update(data)
    return m.hexdigest()


def load_module(mod_name):
    assert isinstance(mod_name, str) and mod_name.find('.') > 0

    cbs = mod_name.split('.')
    pkg, mod = '.'.join(cbs[:-1]), cbs[-1]
    pkg = __import__(pkg, fromlist=[''])
    mod = getattr(pkg, mod)
    return mod


def load_data(data, **kwargs):
    assert isinstance(data, (str, pd.DataFrame, dd.DataFrame))

    if isinstance(data, (pd.DataFrame, dd.DataFrame)):
        return data

    import os.path as path
    import glob

    try:
        from dask.distributed import default_client as dask_default_client
        client = dask_default_client()
        dask_enabled, worker_count = True, len(client.ncores())
    except ValueError:
        dask_enabled, worker_count = False, 1

    fmt_mapping = {
        'csv': (partial(pd.read_csv, low_memory=False), dd.read_csv),
        'txt': (pd.read_csv, dd.read_csv),
        'parquet': (pd.read_parquet, dd.read_parquet),
        'par': (pd.read_parquet, dd.read_parquet),
        'json': (pd.read_json, dd.read_json),
        'pkl': (pd.read_pickle, None),
        'pickle': (pd.read_pickle, None),
    }

    def get_file_format(file_path):
        return path.splitext(file_path)[-1].lstrip('.')

    def get_file_format_by_glob(data_pattern):
        for f in glob.glob(data_pattern, recursive=True):
            fmt_ = get_file_format(f)
            if fmt_ in fmt_mapping.keys():
                return fmt_
        return None

    if glob.has_magic(data):
        fmt = get_file_format_by_glob(data)
    elif not path.exists(data):
        raise ValueError(f'Not found path {data}')
    elif path.isdir(data):
        path_pattern = f'{data}*' if data.endswith(path.sep) else f'{data}{path.sep}*'
        fmt = get_file_format_by_glob(path_pattern)
    else:
        fmt = path.splitext(data)[-1].lstrip('.')

    if fmt not in fmt_mapping.keys():
        # fmt = fmt_mapping.keys()[0]
        raise ValueError(f'Not supported data format{fmt}')
    fn = fmt_mapping[fmt][int(dask_enabled)]
    if fn is None:
        raise ValueError(f'Not supported data format{fmt}')

    if dask_enabled and path.isdir(data) and not glob.has_magic(data):
        data = f'{data}*' if data.endswith(path.sep) else f'{data}{path.sep}*'
    df = fn(data, **kwargs)

    if dask_enabled and worker_count > 1 and df.npartitions < worker_count:
        df = df.repartition(npartitions=worker_count)

    return df



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from .hypergbm import HyperGBMEstimator
from .auto_sklearn import AutoSklearnEstimator
from .h2o import H2OEstimator
from .hyperdt import HyperDTEstimator
from . import Evaluator
from ..datasets import dsutils
from hypergbm.search_space import search_space_feature_gen
import pandas as pd


class Test_Evaluator():
    def test_hypergbm_multiclass(self):
        def get_data(ds_name, task):
            dir = '/Users/jack/workspace/aps/notebook/hypergbm'
            train = pd.read_csv(f'{dir}/datasets/cooka_data/{task}/{ds_name}/train.csv')
            test = pd.read_csv(f'{dir}/datasets/cooka_data/{task}/{ds_name}/test.csv')
            return train, test

        task = 'multiclass'
        target = 'Class'
        X = get_data('yeast', task)
        # X = dsutils.load_glass_uci()
        hypergbm_estimator = HyperGBMEstimator(task=task, scorer='roc_auc_ovo',
                                               eval_size=0.1,
                                               cv=False,
                                               max_trials=3)
        autosklearn_estimator = AutoSklearnEstimator(task=task, time_left_for_this_task=30,
                                                     per_run_time_limit=10)
        h2o_estimator = H2OEstimator(task=task)
        hyperdt_estimator = HyperDTEstimator(task=task, reward_metric='AUC', max_trials=3, epochs=1)
        evaluator = Evaluator()
        result = evaluator.evaluate(X,
                                    target=target,
                                    task='multiclas',
                                    estimators=[
                                        # autosklearn_estimator,
                                        hypergbm_estimator,
                                        # h2o_estimator,
                                        # hyperdt_estimator,
                                    ],

                                    scorers=['roc_auc_ovo'],
                                    test_size=0.3,
                                    random_state=9527)
        assert result

    def test_splitdata(self):
        X = dsutils.load_blood()  # .load_bank().head(1000)
        task = 'binary'
        hypergbm_estimator = HyperGBMEstimator(task=task, scorer='roc_auc_ovo')

        evaluator = Evaluator()
        result = evaluator.evaluate((X[:-100], X[-100:]),
                                    target='Class',
                                    task=task,
                                    estimators=[
                                        # autosklearn_estimator,
                                        hypergbm_estimator,
                                        # h2o_estimator,
                                        # hyperdt_estimator,
                                        # hypergbm_estimator_fg
                                    ],
                                    scorers=['accuracy', 'roc_auc_ovo'],
                                    test_size=0.3,
                                    random_state=9527)

        assert result

    def test_hypergbm_cv_binary(self):
        X = dsutils.load_telescope()  # .load_bank().head(1000)
        task = 'binary'
        hypergbm_estimator = HyperGBMEstimator(task=task, scorer='roc_auc_ovo', cv=True, num_folds=3)
        hypergbm_estimator = HyperGBMEstimator(task=task, scorer='roc_auc_ovo', mode='two-stage', max_trials=300,
                                               earlystop_rounds=10,
                                               use_cache=False,
                                               drop_feature_with_collinearity=False,
                                               ensemble_size=20, use_meta_learner=False, eval_size=0.1,
                                               retrain_on_wholedata=False,
                                               # class_balancing='sample_weight',
                                               cv=True, num_folds=3,
                                               two_stage_importance_selection=False,
                                               pseudo_labeling=False, pseudo_labeling_proba_threshold=0.6,
                                               pseudo_labeling_resplit=False,
                                               )
        evaluator = Evaluator()
        result = evaluator.evaluate(X,
                                    target='Class',
                                    task=task,
                                    estimators=[
                                        hypergbm_estimator
                                    ],
                                    scorers=['roc_auc_ovo'],
                                    test_size=0.3,
                                    random_state=9527)
        assert result

    def test_all_binary(self):
        X = dsutils.load_blood()  # .load_bank().head(1000)
        task = 'binary'
        hypergbm_estimator = HyperGBMEstimator(task=task, scorer='roc_auc_ovo')
        hypergbm_estimator_fg = HyperGBMEstimator(task=task, scorer='roc_auc_ovo', max_trials=3,
                                                  search_space_fn=lambda: search_space_feature_gen(
                                                      early_stopping_rounds=20, verbose=0, task=task))

        autosklearn_estimator = AutoSklearnEstimator(task=task, time_left_for_this_task=30,
                                                     per_run_time_limit=10)
        h2o_estimator = H2OEstimator(task=task)
        hyperdt_estimator = HyperDTEstimator(task=task, reward_metric='AUC', max_trials=3, epochs=1)
        evaluator = Evaluator()
        result = evaluator.evaluate(X,
                                    target='Class',
                                    task=task,
                                    estimators=[
                                        # autosklearn_estimator,
                                        # hypergbm_estimator,
                                        # h2o_estimator,
                                        hyperdt_estimator,
                                        # hypergbm_estimator_fg
                                    ],
                                    scorers=['accuracy', 'roc_auc_ovo'],
                                    test_size=0.3,
                                    random_state=9527)
        assert result



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from hypernets.core.callbacks import EarlyStoppingCallback, EarlyStoppingError
import pytest


class Test_Callback:
    def test_early_stopping(self):
        es = EarlyStoppingCallback(3, 'min')
        es.on_trial_end(None, None, 1, 0.9, True, 0)
        es.on_trial_end(None, None, 2, 0.9, True, 0)
        es.on_trial_end(None, None, 3, 0.9, True, 0)

        with pytest.raises(EarlyStoppingError) as ese:
            es.on_trial_end(None, None, 4, 0.9, True, 0)
        assert ese.value.args[0].find('reason: max_no_improvement_trials') > -0

        es = EarlyStoppingCallback(3, 'min')
        es.on_trial_end(None, None, 1, 0.9, True, 0)
        es.on_trial_end(None, None, 2, 0.8, True, 0)
        es.on_trial_end(None, None, 3, 0.8, True, 0)
        es.on_trial_end(None, None, 4, 0.8, True, 0)

        with pytest.raises(EarlyStoppingError) as ese:
            es.on_trial_end(None, None, 5, 0.8, True, 0)
        assert ese.value.args[0].find('reason: max_no_improvement_trials') > -0

        es = EarlyStoppingCallback(3, 'max')
        es.on_trial_end(None, None, 1, 0.9, True, 0)
        es.on_trial_end(None, None, 2, 0.9, True, 0)
        es.on_trial_end(None, None, 3, 0.9, True, 0)

        with pytest.raises(EarlyStoppingError) as ese:
            es.on_trial_end(None, None, 4, 0.9, True, 0)
        assert ese.value.args[0].find('reason: max_no_improvement_trials') > -0

        es = EarlyStoppingCallback(3, 'max')
        es.on_trial_end(None, None, 1, 0.9, True, 0)
        es.on_trial_end(None, None, 2, 0.91, True, 0)
        es.on_trial_end(None, None, 3, 0.91, True, 0)
        es.on_trial_end(None, None, 4, 0.91, True, 0)

        with pytest.raises(EarlyStoppingError) as ese:
            es.on_trial_end(None, None, 5, 0.91, True, 0)
        assert ese.value.args[0].find('reason: max_no_improvement_trials') > -0



from hypernets.utils import tic_toc, tic_toc_report_as_dataframe
from hypernets.tabular.datasets import dsutils


@tic_toc(details=True)
def fn_foo(a1, a2, k1=None, k2='foo'):
    pass


class ClsBar:
    @tic_toc(details=False)
    def no_args(self):
        pass

    @tic_toc(details=True)
    def method_bar(self, a1, a2, k1=None, k2='foo'):
        pass


def foo():
    fn_foo(1, 2, k1='lalala')
    fn_foo('dict', {'a': 'aaa', 'b': 345})
    fn_foo('list', list(range(5)))
    fn_foo('big-list', list(range(100)))
    fn_foo('big-range', range(100))
    fn_foo('df', dsutils.load_blood())
    fn_foo('ndarray', dsutils.load_blood().values)
    fn_foo('fn', foo)
    fn_foo('lambda', lambda: print('lambda'))
    fn_foo(['aaa', 3, 4, ['aaa', 'bbb']], 2, k2='lalala')


def cls_foo():
    x = ClsBar()
    x.method_bar(1, 2, k1='foo')
    x.method_bar('dict', {'a': 'aaa', 'b': 345})
    x.no_args()


def test_tic_toc():
    foo()
    cls_foo()

    df = tic_toc_report_as_dataframe()
    print(df)



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""

from .base_ensemble import BaseEnsemble
from .stacking import StackingEnsemble
from .voting import AveragingEnsemble, GreedyEnsemble
from .dask_ensemble import DaskGreedyEnsemble


# -*- coding:utf-8 -*-
"""

"""
from hypernets.core.ops import Identity, HyperInput, ConnectionSpace
from hypernets.core.search_space import *


class Test_HyperSpace():
    def get_space(self):
        space = HyperSpace()
        with space.as_default():
            input1 = HyperInput()
            input2 = HyperInput()

            id1 = Identity(p1=Choice([1, 2]), p2=Int(1, 100))
            id2 = Identity(p3=Real(0, 1))
            id3 = Identity(p4=Dynamic(lambda p5: p5 * 3, p5=Choice([2, 4, 8])))
            id4 = Identity()
            id5 = Identity()
            id6 = Identity()
            id7 = Identity()

            id1([input1, id7])
            id2([input1, input2])
            id3([input2, id1, id2])
            id4([input1, id2])
            id5(input2)
            id6(id3)
            id7([id4, id5])
        return space

    def get_space_with_dynamic(self):

        space = HyperSpace()
        with space.as_default():
            input1 = HyperInput()
            input2 = HyperInput()

            id1 = Identity(p1=Choice([1, 2]), p2=Int(1, 100))
            id2 = Identity(p3=Real(0, 1))
            id3 = Identity(p4=Dynamic(lambda p5: p5 * 3, p5=Choice([2, 4, 8])))

            def new_module(m):
                dm1 = Identity(dp1=Choice(['a', 'b']))
                return dm1, dm1

            id4 = ConnectionSpace(dynamic_fn=new_module, p6=Choice(['f', 'g']))
            id5 = Identity()
            id6 = Identity()
            id7 = Identity()

            id1([input1, id7])
            id2([input1, input2])
            id3([input2, id1, id2])
            id4([input1, id2])
            id5(input2)
            id6(id3)
            id7([id4, id5])
        return space

    def test_basic_ops(self):
        with HyperSpace().as_default() as space:
            id1 = HyperInput()
            id2 = HyperInput()
            id3 = Identity()
            id3([id1, id2])
            id4 = Identity()
            id4(id3)
            id5 = Identity()
            id5(id3)

            graph_inputs = space.get_inputs()
            id3_inputs = space.get_inputs(id3)
            id4_inputs = space.get_inputs(id4)

            assert graph_inputs == [id1, id2]
            assert id3_inputs == [id1, id2]
            assert id4_inputs == [id3]

            graph_outputs = space.get_outputs()
            id1_outputs = space.get_outputs(id1)
            id3_outputs = space.get_outputs(id3)

            assert graph_outputs == [id4, id5]
            assert id1_outputs == [id3]
            assert id3_outputs == [id4, id5]

            space.disconnect(id3, id5)
            id3_outputs = space.get_outputs(id3)
            assert id3_outputs == [id4]

            assert len(space.modules) == 5
            assert len(space.edges) == 3

    def test_traverse(self):
        id_list = []

        def print_module(m):
            print(m.id)
            id_list.append(m.id)
            return True

        space = self.get_space()
        space.traverse(print_module, direction='forward')
        assert id_list == ['Module_HyperInput_1', 'Module_HyperInput_2', 'Module_Identity_2', 'Module_Identity_4',
                           'Module_Identity_5', 'Module_Identity_7', 'Module_Identity_1', 'Module_Identity_3',
                           'Module_Identity_6']

        id_list = []
        space.traverse(print_module, direction='backward')
        assert id_list == ['Module_Identity_6', 'Module_Identity_3', 'Module_Identity_1', 'Module_Identity_7',
                           'Module_Identity_4', 'Module_Identity_5', 'Module_Identity_2',
                           'Module_HyperInput_2', 'Module_HyperInput_1', ]

        space = self.get_space_with_dynamic()
        id_list = []
        space.traverse(print_module, direction='forward')
        assert id_list == ['Module_HyperInput_1', 'Module_HyperInput_2', 'Module_Identity_2', 'Module_Identity_4',
                           'Module_ConnectionSpace_1', 'Module_Identity_6', 'Module_Identity_1', 'Module_Identity_3',
                           'Module_Identity_5']

        space.Param_Choice_3.random_sample()
        id_list = []
        space.traverse(print_module, direction='forward')
        assert id_list == ['Module_HyperInput_1', 'Module_HyperInput_2', 'Module_Identity_2',
                           'Module_Identity_7', 'Module_Identity_4', 'Module_Identity_6', 'Module_Identity_1',
                           'Module_Identity_3', 'Module_Identity_5']

    def test_all_assigned(self):
        space = self.get_space()

        assert space.all_assigned == False
        space.Param_Choice_1.random_sample()
        assert space.all_assigned == False
        space.Param_Int_1.random_sample()
        assert space.all_assigned == False
        space.Param_Real_1.assign(0.1)
        assert space.all_assigned == False
        space.Param_Choice_2.random_sample()
        assert space.all_assigned == True

    def test_get_assignable_params(self):
        space = self.get_space()
        with space.as_default():
            const = Constant(1)
        assert const in space.hyper_params
        space.random_sample()
        ps = space.get_assigned_params()
        assert ps == [space.Param_Real_1, space.Param_Choice_1, space.Param_Int_1, space.Param_Choice_2]

    def test_parameter_grid(self):
        from sklearn.model_selection import ParameterGrid
        space = self.get_space()
        ps = space.get_unassigned_params()
        grid = {}
        for p in ps:
            grid[p.name] = [s.value for s in p.expansion(2)]
        all_vectors = list(ParameterGrid(grid))
        for ps in all_vectors:
            space = self.get_space()
            for k,v in ps.items():
                space.__dict__[k].assign(v)
            assert space.all_assigned == True

    def test_params_iterator(self):
        space = self.get_space()
        ps = []
        for p in space.params_iterator:
            ps.append(p)
        assert ps == [space.Param_Real_1, space.Param_Choice_1, space.Param_Int_1, space.Param_Choice_2]
        space.Param_Choice_1.random_sample()

        ps = []
        for p in space.params_iterator:
            ps.append(p)
        assert ps == [space.Param_Real_1, space.Param_Int_1, space.Param_Choice_2]

        space = self.get_space_with_dynamic()
        ps = []
        for p in space.params_iterator:
            ps.append(p)
        assert ps == [space.Param_Real_1, space.Param_Choice_3, space.Param_Choice_1, space.Param_Int_1,
                      space.Param_Choice_2]

        space.Param_Choice_3.random_sample()
        ps = []
        for p in space.params_iterator:
            ps.append(p)
        assert ps == [space.Param_Real_1, space.Param_Choice_4, space.Param_Choice_1, space.Param_Int_1,
                      space.Param_Choice_2]

    def test_random_sample(self):
        space = self.get_space_with_dynamic()
        ps = []
        space.random_sample()
        assert len(space.get_assigned_params()) == 6
        assert len(space.get_all_params()) == 7

        assert space.all_assigned
        for p in space.params_iterator:
            ps.append(p)
        assert ps == []

    def test_default(self):
        dg1 = get_default_space()
        g = HyperSpace()
        with g.as_default():
            dg2 = get_default_space()
            assert dg2 == g
            assert dg2 != dg1

            with HyperSpace().as_default() as g2:
                dg3 = get_default_space()
                assert dg3 == g2
                assert dg3 != dg2

            dg4 = get_default_space()
            assert dg4 == dg2

        dg5 = get_default_space()
        assert dg5 == dg1

    def test_vectors(self):
        space = self.get_space_with_dynamic()
        space.random_sample()
        assert len(space.get_assigned_params()) == 6

        space1 = self.get_space_with_dynamic()
        space1.assign_by_vectors(space.vectors)

        assert space.vectors == space1.vectors

    def test_combinations(self):
        space = self.get_space_with_dynamic()
        assert space.combinations == 119988

    def test_get_subnet_outputs(self):
        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()(id1)
            assert space.get_sub_graph_outputs(id1) == [id2]

        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()
            id3 = Identity()([id1, id2])
            assert space.get_sub_graph_outputs(id1) == [id3]
            assert space.get_sub_graph_outputs(id2) == [id3]

        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()
            id3 = Identity()([id1, id2])
            id4 = Identity()(id3)
            id5 = Identity()(id3)
            assert space.get_sub_graph_outputs(id3) == [id4, id5]
            assert space.get_sub_graph_outputs(id1) == [id4, id5]
            assert space.get_sub_graph_outputs(id2) == [id4, id5]

        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()
            id3 = Identity()([id1, id2])
            id31 = Identity()(id3)
            id32 = Identity()(id31)
            id4 = Identity()(id32)
            id5 = Identity()(id32)
            id11 = Identity()
            id31(id11)
            id41 = Identity()(id3)
            assert space.get_sub_graph_outputs(id1) == [id4, id5, id41]
            assert space.get_sub_graph_outputs(id2) == [id4, id5, id41]
            assert space.get_sub_graph_outputs(id3) == [id4, id5, id41]
            assert space.get_sub_graph_outputs(id31) == [id4, id5, id41]
            assert space.get_sub_graph_outputs(id32) == [id4, id5, id41]
            assert space.get_sub_graph_outputs(id4) == [id4, id5, id41]
            assert space.get_sub_graph_outputs(id5) == [id4, id5, id41]
            assert space.get_sub_graph_outputs(id41) == [id4, id5, id41]

    def test_get_subnet_inputs(self):
        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()(id1)
            assert space.get_sub_graph_inputs(id2) == [id1]

        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()
            id3 = Identity()([id1, id2])
            assert space.get_sub_graph_inputs(id3) == [id1, id2]

        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()(id1)
            id3 = Identity()(id1)
            assert space.get_sub_graph_inputs(id3) == [id1]

        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()(id1)
            id3 = Identity()(id2)
            assert space.get_sub_graph_inputs(id3) == [id1]

        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()(id1)
            id3 = Identity()(id2)
            id4 = Identity()(id3)
            assert space.get_sub_graph_inputs(id3) == [id1]
            assert space.get_sub_graph_inputs(id4) == [id1]

        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()(id1)
            id3 = Identity()
            id4 = Identity()([id2, id3])
            assert space.get_sub_graph_inputs(id4) == [id1, id3]

        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()
            id3 = Identity()([id1, id2])
            id4 = Identity()(id3)
            id5 = Identity()(id3)
            assert space.get_sub_graph_inputs(id3) == [id1, id2]
            assert space.get_sub_graph_inputs(id4) == [id1, id2]
            assert space.get_sub_graph_inputs(id5) == [id1, id2]
            assert space.get_sub_graph_inputs(id1) == [id1, id2]
            assert space.get_sub_graph_inputs(id2) == [id1, id2]

        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()
            id3 = Identity()([id1, id2])
            id31 = Identity()(id3)
            id32 = Identity()(id31)
            id4 = Identity()(id32)
            id5 = Identity()(id32)
            assert space.get_sub_graph_inputs(id3) == [id1, id2]
            assert space.get_sub_graph_inputs(id4) == [id1, id2]
            assert space.get_sub_graph_inputs(id5) == [id1, id2]
            assert space.get_sub_graph_inputs(id1) == [id1, id2]
            assert space.get_sub_graph_inputs(id2) == [id1, id2]

        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()
            id3 = Identity()([id1, id2])
            id31 = Identity()(id3)
            id32 = Identity()(id31)
            id4 = Identity()(id32)
            id5 = Identity()(id32)
            id11 = Identity()
            id31(id11)
            assert space.get_sub_graph_inputs(id3) == [id1, id2, id11]
            assert space.get_sub_graph_inputs(id4) == [id1, id2, id11]
            assert space.get_sub_graph_inputs(id5) == [id1, id2, id11]
            assert space.get_sub_graph_inputs(id1) == [id1, id2, id11]
            assert space.get_sub_graph_inputs(id2) == [id1, id2, id11]

    def test_traverse_subgraph(self):
        space = HyperSpace()
        with space.as_default():
            id1 = Identity()
            id2 = Identity()
            id3 = Identity()([id1, id2])
            id31 = Identity()(id3)
            id32 = Identity()(id31)
            id4 = Identity()(id32)
            id5 = Identity()(id32)
            id11 = Identity()
            id31(id11)
            id_isolated = Identity()
            start = space.get_sub_graph_inputs(id4)
            assert start == [id1, id2, id11]
            id_list = []

            def get_id(m):
                id_list.append(m.id)
                return True

            space.traverse(get_id, direction='forward', start_modules=start)
            assert id_list == ['Module_Identity_1', 'Module_Identity_2', 'Module_Identity_8', 'Module_Identity_3',
                               'Module_Identity_4', 'Module_Identity_5', 'Module_Identity_6', 'Module_Identity_7']

            start = space.get_sub_graph_outputs(id1)
            assert start == [id4, id5]
            id_list = []
            space.traverse(get_id, direction='backward', start_modules=start)
            assert id_list == ['Module_Identity_6', 'Module_Identity_7', 'Module_Identity_5', 'Module_Identity_4',
                               'Module_Identity_3', 'Module_Identity_8', 'Module_Identity_1', 'Module_Identity_2']



from sklearn.model_selection import train_test_split

from hypernets.examples.plain_model import PlainModel, PlainSearchSpace
from hypernets.experiment import GeneralExperiment
from hypernets.searchers import make_searcher
from hypernets.tabular.datasets import dsutils


def create_hyper_model(reward_metric='auc', optimize_direction='max'):
    search_space = PlainSearchSpace()
    searcher = make_searcher('random', search_space_fn=search_space, optimize_direction=optimize_direction)
    hyper_model = PlainModel(searcher=searcher, reward_metric=reward_metric, callbacks=[])

    return hyper_model


def test_general_experiment_of_heart_disease_simple():
    hyper_model = create_hyper_model()

    X = dsutils.load_heart_disease_uci()
    y = X.pop('target')

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3)

    experiment = GeneralExperiment(hyper_model, X_train, y_train, eval_size=0.3)
    estimator = experiment.run(max_trials=5)
    trials = hyper_model.get_top_trials(5)

    assert estimator
    assert 1 < len(trials) <= 5

    score = estimator.evaluate(X_test, y_test, metrics=['auc', 'accuracy', 'f1', 'recall', 'precision'])
    print('evaluate score:', score)
    assert score


def test_general_experiment_of_heart_disease_with_eval_and_cv():
    hyper_model = create_hyper_model()

    X = dsutils.load_heart_disease_uci()
    y = X.pop('target')

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3)
    X_train, X_eval, y_train, y_eval = train_test_split(X_train, y_train, test_size=0.3)

    experiment = GeneralExperiment(hyper_model, X_train, y_train, X_eval=X_eval, y_eval=y_eval, X_test=X_test)
    estimator = experiment.run(max_trials=5, cv=True)
    trials = hyper_model.get_top_trials(5)

    assert estimator
    assert 1 < len(trials) <= 5

    score = estimator.evaluate(X_test, y_test, metrics=['auc', 'accuracy', 'f1', 'recall', 'precision'])
    print('evaluate score:', score)
    assert score



from hypernets.conf import configure, Configurable, Bool, Int, String, List, Dict

try:
    import hn_widget

    default_experiment_ncb = ['hypernets.core.nb_callbacks.JupyterWidgetExperimentCallback', ]
    default_hyper_model_ncb = ['hypernets.core.nb_callbacks.JupyterHyperModelCallback', ]
except ImportError:
    default_experiment_ncb = ['hypernets.experiment.SimpleNotebookCallback', ]
    default_hyper_model_ncb = ['hypernets.core.callbacks.NotebookCallback',
                               'hypernets.core.callbacks.ProgressiveCallback', ]


@configure()
class ExperimentCfg(Configurable):
    experiment_callbacks_console = \
        List(default_value=[],
             allow_none=True, config=True,
             help='ExperimentCallback instance or name list.'
             )
    experiment_callbacks_notebook = \
        List(default_value=default_experiment_ncb,
             allow_none=True, config=True,
             help='ExperimentCallback instance or name list.'
             )

    experiment_default_target_set = \
        List(default_value=['y', 'target', 'class'],
             allow_none=True, config=True,
             help='Default target names.'
             )
    experiment_auto_down_sample_enabled = \
        Bool(False,
             allow_none=True, config=True,
             help=''
             )
    experiment_auto_down_sample_rows_threshold = \
        Int(10000,
            allow_none=True, config=True,
            help=''
            )
    experiment_discriminator = \
        String('percentile',
               allow_none=True, config=True,
               help='discriminator identity, "percentile" or "progressive"',
               )
    experiment_discriminator_options = \
        Dict(default_value={'percentile': 50, 'min_trials': 5, 'min_steps': 5, 'stride': 1},
             key_trait=String,
             allow_none=True, config=True,
             help='discriminator settings',
             )

    hyper_model_callbacks_console = \
        List(default_value=['hypernets.core.callbacks.SummaryCallback', ],
             allow_none=True, config=True,
             help='Callback instance or name list.'
             )
    hyper_model_callbacks_notebook = \
        List(default_value=default_hyper_model_ncb,
             allow_none=True, config=True,
             help='Callback instance or name list.'
             )



# -*- coding:utf-8 -*-
__author__ = 'yangjian'
"""

"""
import numpy as np
from sklearn.inspection import permutation_importance as sk_permutation_importance
from sklearn.utils import Bunch

from hypernets.tabular import dask_ex as dex
from hypernets.utils import logging
from .cfg import TabularCfg as c

logger = logging.get_logger(__name__)

_STRATEGY_THRESHOLD = 'threshold'
_STRATEGY_QUANTILE = 'quantile'
_STRATEGY_NUMBER = 'number'
_STRATEGY_DEFAULT = _STRATEGY_THRESHOLD

_DEFAULT_THRESHOLD = 0.1
_DEFAULT_QUANTILE = 0.2
_DEFAULT_TOP_PERCENT = 0.8


def detect_strategy(strategy=None, threshold=None, quantile=None, number=None):
    if strategy is None:
        if threshold is not None:
            strategy = _STRATEGY_THRESHOLD
        elif number is not None:
            strategy = _STRATEGY_NUMBER
        elif quantile is not None:
            strategy = _STRATEGY_QUANTILE
        else:
            strategy = _STRATEGY_DEFAULT

    if strategy == _STRATEGY_THRESHOLD:
        if threshold is None:
            threshold = _DEFAULT_THRESHOLD
        assert 0 < threshold < 1.0
    elif strategy == _STRATEGY_NUMBER:
        if number is None:
            number = _DEFAULT_TOP_PERCENT
    elif strategy == _STRATEGY_QUANTILE:
        if quantile is None:
            quantile = _DEFAULT_QUANTILE
        assert 0 < quantile < 1.0
    else:
        raise ValueError(f'Unsupported strategy: {strategy}')

    return strategy, threshold, quantile, number


def select_by_feature_importance(feature_importance, strategy=None,
                                 threshold=None, quantile=None, number=None):
    assert isinstance(feature_importance, (list, tuple, np.ndarray)) and len(feature_importance) > 0

    strategy, threshold, quantile, number = detect_strategy(strategy, threshold, quantile, number)
    if strategy == _STRATEGY_NUMBER and isinstance(number, float) and 0 < number < 1.0:
        number = len(feature_importance) * _DEFAULT_TOP_PERCENT

    feature_importance = np.array(feature_importance)
    idx = np.arange(len(feature_importance))

    if strategy == _STRATEGY_THRESHOLD:
        selected = np.where(np.where(feature_importance >= threshold, idx, -1) >= 0)[0]
    elif strategy == _STRATEGY_QUANTILE:
        q = np.quantile(feature_importance, quantile)
        selected = np.where(np.where(feature_importance >= q, idx, -1) >= 0)[0]
    elif strategy == _STRATEGY_NUMBER:
        pos = len(feature_importance) - number
        sorted = np.argsort(np.argsort(feature_importance))
        selected = np.where(sorted >= pos)[0]
    else:
        raise ValueError(f'Unsupported strategy: {strategy}')

    unselected = list(set(range(len(feature_importance))) - set(selected))
    unselected = np.array(unselected)

    return selected, unselected


def permutation_importance_batch(estimators, X, y, scoring=None, n_repeats=5,
                                 n_jobs=None, random_state=None):
    """Evaluate the importance of features of a set of estimators

    Parameters
    ----------
    estimator : list
        A set of estimators that has already been :term:`fitted` and is compatible
        with :term:`scorer`.

    X : ndarray or DataFrame, shape (n_samples, n_features)
        Data on which permutation importance will be computed.

    y : array-like or None, shape (n_samples, ) or (n_samples, n_classes)
        Targets for supervised or `None` for unsupervised.

    scoring : string, callable or None, default=None
        Scorer to use. It can be a single
        string (see :ref:`scoring_parameter`) or a callable (see
        :ref:`scoring`). If None, the estimator's default scorer is used.

    n_repeats : int, default=5
        Number of times to permute a feature.

    n_jobs : int or None, default=None
        The number of jobs to use for the computation.
        `None` means 1 unless in a :obj:`joblib.parallel_backend` context.
        `-1` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.

    random_state : int, RandomState instance, or None, default=None
        Pseudo-random number generator to control the permutations of each
        feature. See :term:`random_state`.

    Returns
    -------
    result : Bunch
        Dictionary-like object, with attributes:

        importances_mean : ndarray, shape (n_features, )
            Mean of feature importance over `n_repeats`.
        importances_std : ndarray, shape (n_features, )
            Standard deviation over `n_repeats`.
        importances : ndarray, shape (n_features, n_repeats)
            Raw permutation importance scores.
    """
    importances = []

    if dex.is_dask_dataframe(X):
        X_shape = dex.compute(X.shape)[0]
        permutation_importance = dex.permutation_importance
    else:
        X_shape = X.shape
        permutation_importance = sk_permutation_importance

    if X_shape[0] > c.permutation_importance_sample_limit:
        if logger.is_info_enabled():
            logger.info(f'{X_shape[0]} rows data found, sample to {c.permutation_importance_sample_limit}')
        frac = c.permutation_importance_sample_limit / X_shape[0]
        X, _, y, _ = dex.train_test_split(X, y, train_size=frac, random_state=random_state)

    if n_jobs is None:
        n_jobs = c.joblib_njobs

    for i, est in enumerate(estimators):
        if logger.is_info_enabled():
            logger.info(f'score permutation importance by estimator {i}/{len(estimators)}')
        importance = permutation_importance(est, X.copy(), y.copy(),
                                            scoring=scoring, n_repeats=n_repeats, n_jobs=n_jobs,
                                            random_state=random_state)
        importances.append(importance.importances)

    importances = np.reshape(np.stack(importances, axis=2), (X.shape[1], -1), 'F')
    bunch = Bunch(importances_mean=np.mean(importances, axis=1),
                  importances_std=np.std(importances, axis=1),
                  importances=importances,
                  columns=X.columns.to_list())
    return bunch



# -*- coding:utf-8 -*-
"""

"""
from ..core.searcher import Searcher, OptimizeDirection


class RandomSearcher(Searcher):
    def __init__(self, space_fn, optimize_direction=OptimizeDirection.Minimize, space_sample_validation_fn=None):
        Searcher.__init__(self, space_fn, optimize_direction, space_sample_validation_fn=space_sample_validation_fn)

    @property
    def parallelizable(self):
        return True

    def sample(self):
        sample = self._sample_and_check(self._random_sample)
        return sample

    def get_best(self):
        raise NotImplementedError

    def update_result(self, space, result):
        pass

    def reset(self):
        raise NotImplementedError

    def export(self):
        raise NotImplementedError
